"""The better gui for arcade

- Improved events, now fully typed
- UIElements are now called Widgets (like everywhere else)
- Widgets render into a FrameBuffer, which supports in memory drawings with less memory usage
- Support for animated widgets
- Texts are now rendered with pyglet, open easier support for text areas with scrolling
- TextArea with scroll support
"""

from collections import defaultdict
from collections.abc import Iterable
from typing import TypeGuard, TypeVar

from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED, EventDispatcher
from pyglet.input import Controller
from pyglet.math import Vec2

import arcade
from arcade.experimental.controller_window import ControllerWindow
from arcade.gui import UIEvent
from arcade.gui.events import (
    UIControllerButtonPressEvent,
    UIControllerButtonReleaseEvent,
    UIControllerConnectEvent,
    UIControllerDisconnectEvent,
    UIControllerDpadEvent,
    UIControllerStickEvent,
    UIControllerTriggerEvent,
    UIKeyPressEvent,
    UIKeyReleaseEvent,
    UIMouseDragEvent,
    UIMouseMovementEvent,
    UIMousePressEvent,
    UIMouseReleaseEvent,
    UIMouseScrollEvent,
    UIOnUpdateEvent,
    UITextInputEvent,
    UITextMotionEvent,
    UITextMotionSelectEvent,
)
from arcade.gui.surface import Surface
from arcade.gui.widgets import UIWidget
from arcade.types import LBWH, AnchorPoint, Point2, Rect

W = TypeVar("W", bound=UIWidget)


class UIManager(EventDispatcher):
    """
    UIManager is the central component within Arcade's GUI system.
    Handles window events, layout process and rendering.

    To process window events, :py:meth:`UIManager.enable()` has to be called,
    which will inject event callbacks for all window events and redirects them
    through the widget tree.

    If used within a view :py:meth:`UIManager.enable()` should be called from
    :py:meth:`View.on_show_view()` and :py:meth:`UIManager.disable()` should be
    called from :py:meth:`View.on_hide_view()`

    Supports `size_hint` to grow/shrink direct children dependent on window size.
    Supports `size_hint_min` to ensure size of direct children (e.g. UIBoxLayout).
    Supports `size_hint_max` to ensure size of direct children (e.g. UIBoxLayout).

    .. code:: py

        class MyView(arcade.View):
            def __init__():
                super().__init__()
                manager = UIManager()

                manager.add(Dummy())

            def on_show_view(self):
                # Set background color
                self.window.background_color = arcade.color.DARK_BLUE_GRAY

                # Enable UIManager when view is shown to catch window events
                self.ui.enable()

            def on_hide_view(self):
                # Disable UIManager when view gets inactive
                self.ui.disable()

            def on_draw():
                self.clear()

                ...

                manager.draw() # draws the UI on screen

    Args:
        window: The window to bind the UIManager to, if None the current window is used.
    """

    _enabled = False
    _pixelated = False
    """Experimental feature to pixelate the UI, all textures will be rendered pixelated,
    which will mostly influence scaled background images.
    This property has to be set right after the UIManager is created."""

    DEFAULT_LAYER = 0
    OVERLAY_LAYER = 10

    def __init__(self, window: arcade.Window | None = None):
        super().__init__()

        self.window = window or arcade.get_window()
        self._surfaces: dict[int, Surface] = {}
        self.children: dict[int, list[UIWidget]] = defaultdict(list)
        self._requires_render = True
        self.camera = arcade.Camera2D()
        self._render_to_surface_camera = arcade.Camera2D()
        # this camera is used for rendering the UI and should not be changed by the user

        self.register_event_type("on_event")

    def add(self, widget: W, *, index=None, layer=DEFAULT_LAYER) -> W:
        """Add a widget to the :class:`UIManager`.

        Added widgets will receive ui events and be rendered.

        By default, the latest added widget will receive ui events first and will
        be rendered on top of others.

        The UIManager supports layered setups, widgets added to a higher layer are
        drawn above lower layers and receive events first.
        The layer 10 is reserved for overlaying components like dropdowns or tooltips.

        Args:
            widget: widget to add
            index: position a widget is added, None has the highest priority
            layer: layer which the widget should be added to, higher layer are above

        """
        if index is None:
            self.children[layer].append(widget)
        else:
            self.children[layer].insert(max(len(self.children), index), widget)
        widget.parent = self
        self.trigger_render()
        return widget

    def remove(self, child: UIWidget):
        """Removes the given widget from UIManager.

        Args:
            child: widget to remove
        """
        for children in self.children.values():
            if child in children:
                children.remove(child)
                child.parent = None
                self.trigger_render()

    def walk_widgets(
        self, *, root: UIWidget | None = None, layer=DEFAULT_LAYER
    ) -> Iterable[UIWidget]:
        """Walks through widget tree, in reverse draw order (most top drawn widget first)

        Args:
            root: root widget to start from, if None, the layer is used
            layer: layer to search, None will search through all layers
        """
        if layer is None:
            layers = sorted(self.children.keys(), reverse=True)
        else:
            layers = [layer]

        for layer in layers:
            children = root.children if root else self.children[layer]
            for child in reversed(children):
                yield from self.walk_widgets(root=child)
                yield child

    def clear(self):
        """Remove all widgets from UIManager"""
        for layer in self.children.values():
            for widget in layer[:]:
                self.remove(widget)

    def get_widgets_at(
        self, pos: Point2, cls: type[W] | type[UIWidget] = UIWidget, layer=DEFAULT_LAYER
    ) -> Iterable[W]:
        """Yields all widgets containing a position, returns first top laying widgets
        which is instance of cls.

        Args:
            pos: Pos within the widget bounds
            cls: class which the widget should be an instance of
            layer: layer to search, None will search through all layers

        Returns: iterator of widgets of given type at position
        """

        def check_type(widget) -> TypeGuard[W]:
            return isinstance(widget, cls)

        for widget in self.walk_widgets(layer=layer):
            if check_type(widget) and widget.rect.point_in_rect(pos):
                yield widget

    def _get_surface(self, layer: int) -> Surface:
        if layer not in self._surfaces:
            if len(self._surfaces) > 2:
                raise Exception("Don't use too much layers!")

            self._surfaces[layer] = Surface(
                size=self.window.get_size(),
                pixel_ratio=self.window.get_pixel_ratio(),
            )
            self._surfaces[layer]._pixelated = self._pixelated

        return self._surfaces[layer]

    def trigger_render(self):
        """Request rendering of all widgets before next draw"""
        self._requires_render = True

    def execute_layout(self):
        """Execute layout process for all widgets.

        This is automatically called during :py:meth:`UIManager.draw()`.
        """
        self._do_layout()

    def _do_layout(self):
        layers = sorted(self.children.keys())
        for layer in layers:
            surface = self._get_surface(layer)
            if not surface:
                raise ValueError("No surface exists for this layer.")
            surface_width, surface_height = surface.size

            for child in self.children[layer]:
                # prepare children, so size_hints are calculated
                child._prepare_layout()

                # actual layout
                if child.size_hint:
                    sh_x, sh_y = child.size_hint
                    nw = surface_width * sh_x if sh_x else None
                    nh = surface_height * sh_y if sh_y else None
                    child.rect = child.rect.resize(nw, nh, anchor=AnchorPoint.BOTTOM_LEFT)

                if child.size_hint_min:
                    shm_w, shm_h = child.size_hint_min
                    child.rect = child.rect.min_size(
                        shm_w or 0, shm_h or 0, anchor=AnchorPoint.BOTTOM_LEFT
                    )

                if child.size_hint_max:
                    shm_w, shm_h = child.size_hint_max
                    child.rect = child.rect.max_size(
                        shm_w or child.width, shm_h or child.height, anchor=AnchorPoint.BOTTOM_LEFT
                    )

                # continue layout process down the tree
                child._do_layout()

    def _do_render(self, force=False):
        layers = sorted(self.children.keys())
        force = force or self._requires_render
        # already reset here, so it can be set again in case during rendering a
        # widget requests a re-rendering like a UILabel updating its font
        self._requires_render = False
        for layer in layers:
            surface = self._get_surface(layer)

            if surface is None:
                raise ValueError("Surface is None for layer, can't render.")

            with surface.activate():
                if force:
                    surface.clear()

                for child in self.children[layer]:
                    child._do_render(surface, force)

    def enable(self) -> None:
        """Registers handler functions (`on_...`) to :py:attr:`arcade.gui.UIElement`

        on_draw is not registered, to provide full control about draw order,
        so it has to be called by the devs themselves.

        Within a view, this method should be called from :py:meth:`arcade.View.on_show_view()`.
        """
        if not self._enabled:
            self._enabled = True

            if isinstance(self.window, ControllerWindow):
                controller_handlers = {
                    self.on_connect,
                    self.on_disconnect,
                    self.on_stick_motion,
                    self.on_trigger_motion,
                    self.on_button_press,
                    self.on_button_release,
                    self.on_dpad_motion,
                }
            else:
                controller_handlers = set()

            self.window.push_handlers(
                self.on_resize,
                self.on_update,
                self.on_mouse_drag,
                self.on_mouse_motion,
                self.on_mouse_press,
                self.on_mouse_release,
                self.on_mouse_scroll,
                self.on_key_press,
                self.on_key_release,
                self.on_text,
                self.on_text_motion,
                self.on_text_motion_select,
                *controller_handlers,
            )

    def disable(self) -> None:
        """Remove handler functions (`on_...`) from :py:attr:`arcade.Window`

        If every :py:class:`arcade.View` uses its own :py:class:`arcade.gui.UIManager`,
        this method should be called in :py:meth:`arcade.View.on_hide_view()`.
        """
        if self._enabled:
            self._enabled = False

            if isinstance(self.window, ControllerWindow):
                controller_handlers = {
                    self.on_connect,
                    self.on_disconnect,
                    self.on_stick_motion,
                    self.on_trigger_motion,
                    self.on_button_press,
                    self.on_button_release,
                    self.on_dpad_motion,
                }
            else:
                controller_handlers = set()

            self.window.remove_handlers(
                self.on_resize,
                self.on_update,
                self.on_mouse_drag,
                self.on_mouse_motion,
                self.on_mouse_press,
                self.on_mouse_release,
                self.on_mouse_scroll,
                self.on_key_press,
                self.on_key_release,
                self.on_text,
                self.on_text_motion,
                self.on_text_motion_select,
                *controller_handlers,
            )

    def on_update(self, time_delta):
        """Dispatches an update event to all widgets in the UIManager."""
        return self.dispatch_ui_event(UIOnUpdateEvent(self, time_delta))

    def draw(self, **kwargs) -> None:
        """Will draw all widgets to the window.

        UIManager caches all rendered widgets into a framebuffer (something like a
        window sized image) and only updates the framebuffer if a widget requests
        rendering via ``trigger_render()``.

        To ensure that the children are positioned properly,
        a layout process is executed before rendering, changes might also trigger a
        re-rendering of all widgets.

        Layouting is a two-step process:
        1. Prepare layout, which prepares children and updates own values
        2. Do layout, which actually sets the position and size of the children
        """
        # Request widgets to prepare for next frame
        self.execute_layout()

        ctx = self.window.ctx
        with ctx.enabled(ctx.BLEND), self._render_to_surface_camera.activate():
            self._do_render()

        # Correct that the ui changes the currently active camera.
        with self.camera.activate():
            # Draw layers
            with ctx.enabled(ctx.BLEND):
                layers = sorted(self.children.keys())
                for layer in layers:
                    self._get_surface(layer).draw()

    def adjust_mouse_coordinates(self, x: float, y: float) -> tuple[float, float]:
        """This method is used, to translate mouse coordinates to coordinates
        respecting the viewport and projection of cameras.

        It uses the internal camera's map_coordinate methods, and should work with
        all transformations possible with the basic orthographic camera.
        """
        x_, y_, *c = self.camera.unproject((x, y))  # convert screen to ui coordinates
        return x_, y_

    def on_event(self, event) -> bool | None:
        """Forwards an event to all widgets in the UIManager."""
        layers = sorted(self.children.keys(), reverse=True)
        for layer in layers:
            for child in reversed(self.children[layer]):
                if child.dispatch_event("on_event", event):
                    # child can consume an event by returning True
                    return EVENT_HANDLED
        return EVENT_UNHANDLED

    def dispatch_ui_event(self, event: UIEvent):
        """Dispatch a UI event to all widgets in the UIManager,
        by triggering py:meth:`on_event()`.

        Args:
            event: The event to dispatch.
        """
        return self.dispatch_event("on_event", event)

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int):
        """Converts mouse motion event to UI event and dispatches it."""
        x_, y_ = self.adjust_mouse_coordinates(x, y)
        return self.dispatch_ui_event(UIMouseMovementEvent(self, round(x_), round(y_), dx, dy))

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int):
        """Converts mouse press event to UI event and dispatches it."""
        x_, y_ = self.adjust_mouse_coordinates(x, y)
        return self.dispatch_ui_event(
            UIMousePressEvent(self, round(x_), round(y_), button, modifiers)
        )

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int):
        """Converts mouse drag event to UI event and dispatches it."""
        x_, y_ = self.adjust_mouse_coordinates(x, y)
        return self.dispatch_ui_event(
            UIMouseDragEvent(self, round(x_), round(y_), dx, dy, buttons, modifiers)
        )

    def on_mouse_release(self, x: int, y: int, button: int, modifiers: int):
        """Converts mouse release event to UI event and dispatches it."""
        x_, y_ = self.adjust_mouse_coordinates(x, y)
        return self.dispatch_ui_event(
            UIMouseReleaseEvent(self, round(x_), round(y_), button, modifiers)
        )

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        """Converts mouse scroll event to UI event and dispatches it."""
        x_, y_ = self.adjust_mouse_coordinates(x, y)
        return self.dispatch_ui_event(
            UIMouseScrollEvent(self, round(x_), round(y_), scroll_x, scroll_y)
        )

    def on_key_press(self, symbol: int, modifiers: int):
        """Converts key press event to UI event and dispatches it."""
        return self.dispatch_ui_event(UIKeyPressEvent(self, symbol, modifiers))

    def on_key_release(self, symbol: int, modifiers: int):
        """Converts key release event to UI event and dispatches it."""
        return self.dispatch_ui_event(UIKeyReleaseEvent(self, symbol, modifiers))

    def on_text(self, text):
        """Converts text event to UI event and dispatches it."""
        return self.dispatch_ui_event(UITextInputEvent(self, text))

    def on_text_motion(self, motion):
        """Converts text motion event to UI event and dispatches it."""
        return self.dispatch_ui_event(UITextMotionEvent(self, motion))

    def on_text_motion_select(self, motion):
        """Converts text motion select event to UI event and dispatches it."""
        return self.dispatch_ui_event(UITextMotionSelectEvent(self, motion))

    def on_resize(self, width, height):
        """Resize the UIManager and all of its surfaces."""
        # resize ui camera
        bottom_left = self.camera.bottom_left
        self.camera.match_window()
        self.camera.bottom_left = bottom_left

        # resize render to surface camera
        bottom_left = self._render_to_surface_camera.bottom_left
        self._render_to_surface_camera.match_window()
        self._render_to_surface_camera.bottom_left = bottom_left

        scale = self.window.get_pixel_ratio()
        for surface in self._surfaces.values():
            surface.resize(size=(width, height), pixel_ratio=scale)

        self.trigger_render()

    def on_connect(self, controller: Controller):
        """Called when a controller is connected."""
        self.dispatch_ui_event(UIControllerConnectEvent(controller))

    def on_disconnect(self, controller: Controller):
        """Called when a controller is disconnected."""
        self.dispatch_ui_event(UIControllerDisconnectEvent(controller))

    def on_stick_motion(self, controller: Controller, name: str, value: Vec2):
        return self.dispatch_ui_event(UIControllerStickEvent(controller, name, value))

    def on_trigger_motion(self, controller: Controller, name: str, value: float):
        return self.dispatch_ui_event(UIControllerTriggerEvent(controller, name, value))

    def on_button_press(self, controller: Controller, button: str):
        return self.dispatch_ui_event(UIControllerButtonPressEvent(controller, button))

    def on_button_release(self, controller: Controller, button: str):
        return self.dispatch_ui_event(UIControllerButtonReleaseEvent(controller, button))

    def on_dpad_motion(self, controller: Controller, value: Vec2):
        return self.dispatch_ui_event(UIControllerDpadEvent(controller, value))

    @property
    def rect(self) -> Rect:
        """The rect of the UIManager, which is the window size."""
        return LBWH(0, 0, *self.window.get_size())

    def debug(self):
        """Walks through all widgets of a UIManager and prints out layout information."""
        for index, layer in self.children.items():
            print(f"Layer {index}")
            # print table headers, widget, rect, size_hint, size_hint_min, size_hint_max
            print(
                f"{'Widget':<60}|"
                f"{'Rect (LBWH)':<30}|"
                f"{'Size Hint':<12}|"
                f"{'Size Hint Min':<12}|"
                f"{'Size Hint Max':<12}"
            )
            for child in reversed(layer):
                self._debug(child, prefix="  ")
        return

    @staticmethod
    def _debug(element, prefix=""):
        print(
            f"{prefix}{element}".ljust(60),
            f"{tuple(round(v, 2) for v in element.rect.lbwh)}".ljust(30),
            f"{element.size_hint}".ljust(12),
            f"{element.size_hint_min}".ljust(12),
            f"{element.size_hint_max}".ljust(12),
            sep="|",
        )
        if isinstance(element, UIWidget):
            for child in element.children:
                UIManager._debug(child, prefix=prefix + "  ")



import array

from arcade import gl
from arcade.types import Color, RGBOrA255
from arcade.window_commands import get_window


def draw_circle_filled(
    center_x: float,
    center_y: float,
    radius: float,
    color: RGBOrA255,
    tilt_angle: float = 0,
    num_segments: int = -1,
) -> None:
    """
    Draw a filled-in circle.

    Args:
        center_x:
            x position that is the center of the circle.
        center_y:
            y position that is the center of the circle.
        radius:
            width of the circle.
        color:
            A 3 or 4 length tuple of 0-255 channel values
            or a :py:class:`.Color` instance.
        tilt_angle:
            Angle in degrees to tilt the circle. Useful for low segment count circles
        num_segments:
            Number of triangle segments that make up this circle.
            Higher is better quality, but slower render time.
            The default value of -1 means Arcade will try to calculate a reasonable
            amount of segments based on the size of the circle.
    """
    draw_ellipse_filled(
        center_x,
        center_y,
        radius * 2,
        radius * 2,
        color,
        tilt_angle=tilt_angle,
        num_segments=num_segments,
    )


def draw_circle_outline(
    center_x: float,
    center_y: float,
    radius: float,
    color: RGBOrA255,
    border_width: float = 1,
    tilt_angle: float = 0,
    num_segments: int = -1,
) -> None:
    """
    Draw the outline of a circle.

    Args:
        center_x:
            x position that is the center of the circle.
        center_y:
            y position that is the center of the circle.
        radius:
            width of the circle.
        color:
            A 3 or 4 length tuple of 0-255 channel values
            or a :py:class:`.Color` instance.
        border_width:
            Width of the circle outline in pixels.
        tilt_angle:
            Angle in degrees to tilt the circle (clockwise).
            Useful for low segment count circles
        num_segments:
            Number of triangle segments that make up this circle.
            Higher is better quality, but slower render time.
            The default value of -1 means Arcade will try to calculate a reasonable
            amount of segments based on the size of the circle.
    """
    draw_ellipse_outline(
        center_x=center_x,
        center_y=center_y,
        width=radius * 2,
        height=radius * 2,
        color=color,
        border_width=border_width,
        tilt_angle=tilt_angle,
        num_segments=num_segments,
    )


def draw_ellipse_filled(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBOrA255,
    tilt_angle: float = 0,
    num_segments: int = -1,
) -> None:
    """
    Draw a filled in ellipse.

    Args:
        center_x:
            x position that is the center of the circle.
        center_y:
            y position that is the center of the circle.
        width:
            width of the ellipse.
        height:
            height of the ellipse.
        color:
            A 3 or 4 length tuple of 0-255 channel values
            or a :py:class:`.Color` instance.
        color:
            Either a :py:class:`.Color` instance
            or an RGBA :py:class:`tuple` of 4 byte values (0 to 255).
        tilt_angle:
            Angle in degrees to tilt the ellipse (clockwise).
            Useful when drawing a circle with a low segment count, to make an octagon for example.
        num_segments:
            Number of triangle segments that make up this circle.
            Higher is better quality, but slower render time.
            The default value of -1 means Arcade will try to calculate a reasonable
            amount of segments based on the size of the circle.
    """
    # Fail immediately if we have no window or context
    window = get_window()
    ctx = window.ctx
    ctx.enable(ctx.BLEND)

    program = ctx.shape_ellipse_filled_unbuffered_program
    geometry = ctx.shape_ellipse_unbuffered_geometry
    buffer = ctx.shape_ellipse_unbuffered_buffer  # type: ignore

    # Normalize the color because this shader takes a float uniform
    color_normalized = Color.from_iterable(color).normalized

    # Pass data to the shader
    program["color"] = color_normalized
    program["shape"] = width / 2, height / 2, tilt_angle
    program["segments"] = num_segments
    buffer.orphan()
    buffer.write(data=array.array("f", (center_x, center_y)))

    geometry.render(program, mode=gl.POINTS, vertices=1)
    ctx.disable(ctx.BLEND)


def draw_ellipse_outline(
    center_x: float,
    center_y: float,
    width: float,
    height: float,
    color: RGBOrA255,
    border_width: float = 1,
    tilt_angle: float = 0,
    num_segments: int = -1,
) -> None:
    """
    Draw the outline of an ellipse.

    Args:
        center_x:
            x position that is the center of the circle.
        center_y:
            y position that is the center of the circle.
        width:
            width of the ellipse.
        height:
            height of the ellipse.
        color:
            A 3 or 4 length tuple of 0-255 channel values
            or a :py:class:`.Color` instance.
        border_width:
            Width of the circle outline in pixels.
        tilt_angle:
            Angle in degrees to tilt the ellipse (clockwise).
            Useful when drawing a circle with a low segment count, to make an octagon for example.
        num_segments:
            Number of triangle segments that make up this circle.
            Higher is better quality, but slower render time.
            The default value of -1 means Arcade will try to calculate a reasonable
            amount of segments based on the size of the circle.
    """
    # Fail immediately if we have no window or context
    window = get_window()
    ctx = window.ctx
    program = ctx.shape_ellipse_outline_unbuffered_program
    geometry = ctx.shape_ellipse_outline_unbuffered_geometry
    buffer = ctx.shape_ellipse_outline_unbuffered_buffer  # type: ignore

    # Normalize the color because this shader takes a float uniform
    color_normalized = Color.from_iterable(color).normalized

    ctx.enable(ctx.BLEND)

    # Pass data to shader
    program["color"] = color_normalized
    program["shape"] = width / 2, height / 2, tilt_angle, border_width
    program["segments"] = num_segments
    buffer.orphan()
    buffer.write(data=array.array("f", (center_x, center_y)))

    geometry.render(program, mode=gl.POINTS, vertices=1)

    ctx.disable(ctx.BLEND)



"""
Box is the 3D counterpart to :py:class:`~arcade.types.rect.Rect`.
"""

from __future__ import annotations

from typing import Any, NamedTuple, TypedDict

from pyglet.math import Vec3

from arcade.types.numbers import AsFloat
from arcade.types.rect import LBWH, Rect
from arcade.types.vector_like import Point3


class BoxKwargs(TypedDict):
    """Annotates a plain :py:class:`dict` of :py:class:`Box` arguments.

    This is only meaningful as a type annotation during type checking.
    For example, the :py:meth:`Box.kwargs <arcade.types.Box.kwargs>`
    property returns an ordinary will actually be a :py:class:`dict`
    of :py:class:`Box` field names to :py:class:`float` values.

    To learn more, please see:

    * :py:class:`.Box`
    * :py:class:`typing.TypedDict`

    """

    left: float
    right: float
    bottom: float
    top: float
    near: float
    far: float
    width: float
    height: float
    depth: float
    x: float
    y: float
    z: float


class Box(NamedTuple):
    """A 3D box, with several convenience properties and functions.

    .. important:: Boxes are immutable and axis-aligned bounding prisms!

    As :py:class:`~typing.NamedTuple` subclasses, they cannot rotate and
    have no setters. This keeps their design simple and efficient. To
    rotate a box's points in 3D, use :py:meth:`to_points` with a 3D math
    library of your choice.

    To create a box, the helper class methods are usually the best choice:

    * :py:func:`.XYZWHD`
    * :py:func:`.LRBTNF`
    * :py:func:`.LBNWHD`
    * :py:meth:`.from_kwargs`

    """

    left: float
    right: float
    bottom: float

    top: float

    near: float
    far: float

    width: float
    height: float
    depth: float

    x: float
    y: float
    z: float

    @property
    def center_x(self) -> float:
        """Backwards-compatible alias for :py:attr:`.x`."""
        return self.x

    @property
    def center_y(self) -> float:
        """Backwards-compatible alias for :py:attr:`.y`."""
        return self.y

    @property
    def center_z(self) -> float:
        """Backwards-compatible alias for :py:attr:`.z`."""
        return self.z

    @property
    def center(self) -> Vec3:
        """Returns a :py:class:`~pyglet.math.Vec3` representing the center of the box."""
        return Vec3(self.x, self.y, self.z)

    @property
    def bottom_left_near(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        bottom-left-near corner of the box.
        """
        return Vec3(self.left, self.bottom, self.near)

    @property
    def bottom_left_far(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        bottom-left-far corner of the box.
        """
        return Vec3(self.left, self.bottom, self.far)

    @property
    def bottom_right_near(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        bottom-right-near corner of the box.
        """
        return Vec3(self.right, self.bottom, self.near)

    @property
    def bottom_right_far(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        bottom-right-far corner of the box.
        """
        return Vec3(self.right, self.bottom, self.far)

    @property
    def top_left_near(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        top-left-near corner of the box.
        """
        return Vec3(self.left, self.top, self.near)

    @property
    def top_left_far(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        top-left-far corner of the box.
        """
        return Vec3(self.left, self.top, self.far)

    @property
    def top_right_near(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        top-right-near corner of the box.
        """
        return Vec3(self.right, self.top, self.near)

    @property
    def top_right_far(self) -> Vec3:
        """
        Returns a :py:class:`~pyglet.math.Vec3` representing the
        top-right-far corner of the box.
        """
        return Vec3(self.right, self.top, self.far)

    @property
    def near_face(self) -> Rect:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        near face of the box.
        """
        return LBWH(self.left, self.bottom, self.width, self.height)

    @property
    def far_face(self) -> Rect:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        far face of the box.
        """
        return LBWH(self.left, self.bottom, self.width, self.height)

    @property
    def left_face(self) -> Rect:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        left face of the box.
        """
        return LBWH(self.near, self.bottom, self.depth, self.height)

    @property
    def right_face(self) -> Rect:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        right face of the box.
        """
        return LBWH(self.near, self.bottom, self.depth, self.height)

    @property
    def top_face(self) -> Rect:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        top face of the box.
        """
        return LBWH(self.left, self.near, self.width, self.depth)

    @property
    def bottom_face(self) -> Rect:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        bottom face of the box.
        """
        return LBWH(self.left, self.near, self.width, self.depth)

    @property
    def near_face_center(self) -> Vec3:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        center of the near face of the box.
        """
        return Vec3(self.x, self.y, self.near)

    @property
    def far_face_center(self) -> Vec3:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        center of the far face of the box.
        """
        return Vec3(self.x, self.y, self.far)

    @property
    def left_face_center(self) -> Vec3:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        center of the left face of the box.
        """
        return Vec3(self.left, self.y, self.z)

    @property
    def right_face_center(self) -> Vec3:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        center of the right face of the box.
        """
        return Vec3(self.right, self.y, self.z)

    @property
    def top_face_center(self) -> Vec3:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        center of the top face of the box.
        """
        return Vec3(self.x, self.top, self.z)

    @property
    def bottom_face_center(self) -> Vec3:
        """
        Returns a :py:class:`~arcade.Rect` representing the
        center of the bottom face of the box.
        """
        return Vec3(self.x, self.top, self.z)

    @property
    def size(self) -> Vec3:
        """Returns a :py:class:`~pyglet.math.Vec3` representing the size of the box."""
        return Vec3(self.width, self.height, self.depth)

    @property
    def volume(self) -> float:
        """The volume of the box in cubic pixels."""
        return self.width * self.height * self.depth

    def at_position(self, position: Point3) -> Box:
        """Returns a new :py:class:`Box` which is moved to put `position` at its center."""
        x, y, z = position
        return XYZWHD(x, y, z, self.width, self.height, self.depth)

    def move(self, dx: AsFloat = 0.0, dy: AsFloat = 0.0, dz: AsFloat = 0.0) -> Box:
        """
        Returns a new :py:class:`Box` which is moved by `dx` in the
        x-direction,`dy` in the y-direction, and `dz` in the z-direction.
        """
        return XYZWHD(self.x + dx, self.y + dy, self.z + dz, self.width, self.height, self.depth)

    def union(self, other: Box) -> Box:
        """Get the smallest Box encapsulating both this one and ``other``."""
        left = min(self.left, other.left)
        right = max(self.right, other.right)
        bottom = min(self.bottom, other.bottom)
        top = max(self.top, other.top)
        near = min(self.near, other.near)
        far = max(self.far, other.far)
        return LRBTNF(left, right, bottom, top, near, far)

    def __or__(self, other: Box) -> Box:
        """Shorthand for :py:meth:`Box.union(other) <union>`.

        Args:
            other: Another :py:class:`Box` instance.
        """
        return self.union(other)

    def intersection(self, other: Box) -> Box | None:
        """Return a :py:class:`Box` of the overlap if any exists.

        If the two :py:class:`Box` instances do not intersect, this
        method will return ``None`` instead.

        Args:
            other: Another :py:class:`Box` instance.
        """
        intersecting = self.overlaps(other)
        if not intersecting:
            return None
        left = max(self.left, other.left)
        right = min(self.right, other.right)
        bottom = max(self.bottom, other.bottom)
        top = min(self.top, other.top)
        near = max(self.near, other.near)
        far = min(self.far, other.far)
        return LRBTNF(left, right, bottom, top, near, far)

    def overlaps(self, other: Box) -> bool:
        """Returns ``True`` if `other` overlaps with ``self``.

        Args:
            other: Another :py:class:`Box` instance.
        """

        return (
            (other.width + self.width) / 2.0 > abs(self.x - other.x)
            and (other.height + self.height) / 2.0 > abs(self.y - other.y)
            and (other.depth + self.depth) / 2.0 > abs(self.z - other.z)
        )

    def __and__(self, other: Box) -> Box | None:
        """Shorthand for :py:meth:`Box.intersection(other) <interesection>`.

        Args:
            other: Another :py:class:`Box` instance.
        """
        return self.intersection(other)

    def point_in_box(self, point: Point3) -> bool:
        """``True`` if the point is in or touching the box, otherwise ``False``.

        Args:
             point: A 3D point.
        """
        x, y, z = point
        return (
            (self.left <= x <= self.right)
            and (self.bottom <= y <= self.top)
            and (self.near <= z <= self.far)
        )

    def __contains__(self, point: Point3 | Any) -> bool:
        """Shorthand for :py:meth:`Box.point_in_box(point) <point_in_box>`.

        Args:
            point: A tuple of :py:class:`int` or :py:class:`float` values.
        """
        from arcade.utils import is_iterable

        if not is_iterable(point):
            return False

        return self.point_in_box(point)

    def to_points(self) -> tuple[Vec3, Vec3, Vec3, Vec3, Vec3, Vec3, Vec3, Vec3]:
        """Return a new :py:class:`tuple` of this box's corners as 3D points.

        The points will be ordered as follows:

        #. :py:meth:`bottom_left_near`
        #. :py:meth:`top_left_near`
        #. :py:meth:`top_right_near`
        #. :py:meth:`bottom_right_near`
        #. :py:meth:`bottom_left_far`
        #. :py:meth:`top_left_far`
        #. :py:meth:`top_right_far`
        #. :py:meth:`bottom_right_far`

        """
        left = self.left
        bottom = self.bottom
        right = self.right
        top = self.top
        near = self.near
        far = self.far
        return (
            Vec3(left, bottom, near),
            Vec3(left, top, near),
            Vec3(right, top, near),
            Vec3(right, bottom, near),
            Vec3(left, bottom, far),
            Vec3(left, top, far),
            Vec3(right, top, far),
            Vec3(right, bottom, far),
        )

    @property
    def xyzwhd(self) -> tuple[AsFloat, AsFloat, AsFloat, AsFloat, AsFloat, AsFloat]:
        """Provides a tuple in the form (x, y, z, width, height, depth)."""
        return (self.x, self.y, self.z, self.width, self.height, self.depth)

    @property
    def lrbtnf(self) -> tuple[AsFloat, AsFloat, AsFloat, AsFloat, AsFloat, AsFloat]:
        """Provides a tuple in the form (left, right, bottom, top, near, far)."""
        return (self.left, self.right, self.bottom, self.top, self.near, self.far)

    @property
    def lbnwhd(self) -> tuple[AsFloat, AsFloat, AsFloat, AsFloat, AsFloat, AsFloat]:
        """Provides a tuple in the form (left, bottom, near, width, height, depth)."""
        return (self.left, self.bottom, self.near, self.width, self.height, self.depth)

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__name__} LRBTNF({self.left}, {self.right},"
            f"{self.bottom}, {self.top}, {self.near}, {self.far})"
            f" XYZWHD({self.x}, {self.y}, {self.z} {self.width}, {self.height}, {self.depth})>"
        )

    def __bool__(self) -> bool:
        """Returns ``True`` if volume is not ``0``, else ``False``."""
        return self.width != 0 and self.height != 0 and self.depth != 0


def XYZWHD(x: float, y: float, z: float, width: float, height: float, depth: float) -> Box:
    """Creates a new :py:class:`.Box` from center x, center y, center z,
    width, height, and depth parameters."""
    h_width = width / 2
    h_height = height / 2
    h_depth = depth / 2
    return Box(
        x - h_width,
        x + h_width,
        y - h_height,
        y + h_height,
        z - h_depth,
        z + h_depth,
        width,
        height,
        depth,
        x,
        y,
        z,
    )


def LBNWHD(
    left: float, bottom: float, near: float, width: float, height: float, depth: float
) -> Box:
    """Creates a new :py:class:`.Box` from left, bottom, near,
    width, height, and depth parameters."""
    return Box(
        left,
        left + width,
        bottom,
        bottom + height,
        near,
        near + depth,
        width,
        height,
        depth,
        left + (width / 2),
        bottom + (height / 2),
        near + (depth / 2),
    )


def LRBTNF(left: float, right: float, bottom: float, top: float, near: float, far: float) -> Box:
    """Creates a new :py:class:`.Box` from left, right, bottom, top, near, and far parameters."""
    width = right - left
    height = top - bottom
    depth = far - near
    return Box(
        left,
        right,
        bottom,
        top,
        near,
        far,
        width,
        height,
        depth,
        left + width / 2.0,
        bottom + height / 2.0,
        near + depth / 2.0,
    )



"""Show all arcade.uicolors in a grid.

Click on a color to select
it and copy the Arcade reference to the clipboard.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.5_uicolor_picker

"""

from dataclasses import dataclass

import arcade
from arcade.gui import (
    UIAnchorLayout,
    UIBoxLayout,
    UIEvent,
    UIGridLayout,
    UIInteractiveWidget,
    UILabel,
    UITextWidget,
    UIView,
)


@dataclass
class ChooseColorEvent(UIEvent):
    """Custom event, which is dispatched when a color button is clicked."""

    color_name: str
    color: arcade.color.Color


class Toast(UILabel):
    """Label which disappears after a certain time."""

    def __init__(self, text: str, duration: float = 2.0, **kwargs):
        super().__init__(**kwargs)
        self.text = text
        self.duration = duration
        self.time = 0

    def on_update(self, dt):
        self.time += dt

        if self.time > self.duration:
            self.parent.remove(self)


class ColorButton(UITextWidget, UIInteractiveWidget):
    """Button which shows a color and color name and
    emits a ChooseColorEvent event when clicked."""

    def __init__(
        self,
        color_name: str,
        color: arcade.color.Color,
        **kwargs,
    ):
        super().__init__(text=color_name, **kwargs)
        # set color and place text on the bottom
        self.with_background(color=color)
        self.place_text(anchor_y="bottom")

        # set font color based on background color
        f = 2 if color_name.startswith("DARK") else 0.5
        self.ui_label.update_font(
            font_color=arcade.color.Color(int(color[0] * f), int(color[1] * f), int(color[2] * f))
        )

        # store color name and color for later reference
        self._color_name = color_name
        self._color = color

        # register custom event
        self.register_event_type("on_choose_color")

    def on_update(self, dt):
        """Update the button state.

        UIInteractiveWidget provides properties like hovered and pressed,
        which can be used to highlight the button."""
        if self.pressed:
            self.with_border(color=arcade.uicolor.WHITE_CLOUDS, width=3)
        elif self.hovered:
            self.with_border(color=arcade.uicolor.WHITE_CLOUDS, width=2)
        else:
            self.with_border(color=arcade.color.BLACK, width=1)

    def on_click(self, event) -> bool:
        """Emit a ChooseColorEvent event when clicked."""
        self.dispatch_event(
            "on_choose_color", ChooseColorEvent(self, self._color_name, self._color)
        )
        return True

    def on_choose_color(self, event: ChooseColorEvent):
        """ChooseColorEvent event handler, which can be overridden."""
        pass


class ColorView(UIView):
    """Uses the arcade.gui.UIView which takes care about the UIManager setup."""

    def __init__(self):
        super().__init__()
        # Create an anchor layout, which can be used to position widgets on screen
        self.root = self.add_widget(UIAnchorLayout())

        # Define colors in grid order
        self.colors = {
            # row 0
            "GREEN_TURQUOISE": arcade.uicolor.GREEN_TURQUOISE,
            "GREEN_EMERALD": arcade.uicolor.GREEN_EMERALD,
            "BLUE_PETER_RIVER": arcade.uicolor.BLUE_PETER_RIVER,
            "PURPLE_AMETHYST": arcade.uicolor.PURPLE_AMETHYST,
            "DARK_BLUE_WET_ASPHALT": arcade.uicolor.DARK_BLUE_WET_ASPHALT,
            # row 1
            "GREEN_GREEN_SEA": arcade.uicolor.GREEN_GREEN_SEA,
            "GREEN_NEPHRITIS": arcade.uicolor.GREEN_NEPHRITIS,
            "BLUE_BELIZE_HOLE": arcade.uicolor.BLUE_BELIZE_HOLE,
            "PURPLE_WISTERIA": arcade.uicolor.PURPLE_WISTERIA,
            "DARK_BLUE_MIDNIGHT_BLUE": arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE,
            # row 2
            "YELLOW_SUN_FLOWER": arcade.uicolor.YELLOW_SUN_FLOWER,
            "ORANGE_CARROT": arcade.uicolor.ORANGE_CARROT,
            "RED_ALIZARIN": arcade.uicolor.RED_ALIZARIN,
            "WHITE_CLOUDS": arcade.uicolor.WHITE_CLOUDS,
            "GRAY_CONCRETE": arcade.uicolor.GRAY_CONCRETE,
            # row 3
            "YELLOW_ORANGE": arcade.uicolor.YELLOW_ORANGE,
            "ORANGE_PUMPKIN": arcade.uicolor.ORANGE_PUMPKIN,
            "RED_POMEGRANATE": arcade.uicolor.RED_POMEGRANATE,
            "WHITE_SILVER": arcade.uicolor.WHITE_SILVER,
            "GRAY_ASBESTOS": arcade.uicolor.GRAY_ASBESTOS,
        }

        # setup grid with colors
        self.grid = self.root.add(
            UIGridLayout(
                column_count=5,
                row_count=4,
                size_hint=(1, 1),
            )
        )
        for i, (name, color) in enumerate(self.colors.items()):
            button = ColorButton(
                color_name=name,
                color=color,
                size_hint=(1, 1),
            )
            self.grid.add(button, row=i // 5, column=i % 5)

            # connect event handler
            button.on_choose_color = self.on_color_button_choose_color

        # setup toasts (temporary messages)
        self.toasts = self.root.add(UIBoxLayout(space_between=2), anchor_x="right", anchor_y="top")
        self.toasts.with_padding(all=10)

    def on_color_button_choose_color(self, event: ChooseColorEvent) -> bool:
        """Color button click event handler, which copies the color name to the clipboard.

        And shows a temporary message."""
        self.window.set_clipboard_text(f"arcade.uicolor.{event.color_name}")

        # prepare and show toast
        toast = Toast(f"Copied {event.color_name}", width=250, size_hint=(None, 0))
        toast.update_font(
            font_color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE,
            font_size=9,
            bold=True,
        )
        toast.with_background(color=arcade.uicolor.GREEN_EMERALD)
        toast.with_padding(all=10)

        self.toasts.add(toast)

        return True

    def on_draw_before_ui(self):
        # Add draw commands that should be below the UI
        pass

    def on_draw_after_ui(self):
        # Add draw commands that should be on top of the UI (uncommon)
        pass


def main():
    window = arcade.Window(title="GUI Example: Color Picker")
    window.show_view(ColorView())
    window.run()


if __name__ == "__main__":
    main()



# ruff: noqa: F401
#  type: ignore

from .inputs import (
    ControllerAxes,
    ControllerButtons,
    XBoxControllerButtons,
    PSControllerButtons,
    Keys,
    MouseAxes,
    MouseButtons,
)
from .manager import ActionState, InputManager
from .input_mapping import Action, ActionMapping, Axis, AxisMapping

__all__ = [
    "ControllerAxes",
    "ControllerButtons",
    "XBoxControllerButtons",
    "PSControllerButtons",
    "Keys",
    "MouseAxes",
    "MouseButtons",
    "ActionState",
    "InputManager",
    "Action",
    "ActionMapping",
    "Axis",
    "AxisMapping",
]



import pyglet.input
from pyglet.input import Joystick

__all__ = ["get_joysticks", "get_game_controllers"]


def get_joysticks() -> list[Joystick]:
    """
    Get a list of all the game controllers

    This is an alias of :func:`get_game_controllers`, which is better worded.
    """
    return pyglet.input.get_joysticks()  # type: ignore  # pending https://github.com/pyglet/pyglet/issues/842


def get_game_controllers() -> list[Joystick]:
    """
    Get a list of all the game controllers
    """
    return get_joysticks()



#  type: ignore
import random
from collections.abc import Sequence

import pyglet
from pyglet.input import Controller

import arcade
from arcade.future.input import ActionState, ControllerAxes, ControllerButtons, InputManager, Keys

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
FLOAT_HEIGHT = 80


# Default per-player textures
DEFAULT_TEXTURES = (
    ":resources:images/animated_characters/female_adventurer/femaleAdventurer_idle.png",
    ":resources:images/animated_characters/male_adventurer/maleAdventurer_idle.png",
    ":resources:images/animated_characters/robot/robot_idle.png",
    ":resources:images/animated_characters/zombie/zombie_idle.png",
)


class Player(arcade.Sprite):
    def __init__(
        self,
        texture,
        walls: arcade.SpriteSequence[arcade.BasicSprite],
        input_manager_template: InputManager,
        controller: pyglet.input.Controller | None = None,
        center_x: float = 0.0,
        center_y: float = 0.0,
        x_max_speed: float = 300.0,
        y_jump_speed: float = 20.0,
    ):
        super().__init__(texture, center_x=center_x, center_y=center_y)
        self.x_max_velocity = x_max_speed
        self.y_jump_speed = y_jump_speed

        self.input_manager = InputManager(controller=controller, action_handlers=self.on_action)
        self.input_manager.copy_existing(input_manager_template)

        #
        self.physics_engine = arcade.PhysicsEnginePlatformer(self, walls=walls, gravity_constant=1)

    def on_update(self, delta_time: float = 1 / 60) -> None:
        self.input_manager.update()
        self.change_x = self.input_manager.axis("Move") * self.x_max_velocity * delta_time
        self.physics_engine.update()

    def on_action(self, action: str, state: ActionState):
        if action == "Jump" and state == ActionState.PRESSED and self.physics_engine.can_jump():
            self.change_y = self.y_jump_speed


class Game(arcade.Window):
    def __init__(
        self,
        player_textures: Sequence[str] = DEFAULT_TEXTURES,
        max_players: int = 4,
        randomize_textures: bool = True,
    ):
        if not (1 < max_players <= 4):
            raise ValueError("max_players must between 1 and 4 (the maximum supported by XInput)")

        super().__init__(title="Input Example")

        self.player_textures = [t for t in player_textures]
        if randomize_textures:
            random.shuffle(self.player_textures)

        self._max_players = max_players
        self.key_to_player_index: dict[Keys, int] = {
            getattr(Keys, f"KEY_{i + 1}"): i for i in range(0, max_players)
        }

        self.players: list[Player | None] = []
        self.player_list: arcade.SpriteList[Player] = arcade.SpriteList()
        self.device_labels_batch = pyglet.graphics.Batch()
        self.player_device_labels: list[arcade.Text | None] = []

        self.wall_list: arcade.SpriteList[arcade.Sprite] = arcade.SpriteList(use_spatial_hash=True)

        for x in range(0, self.width + 64, 64):
            wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=0.5)
            wall.center_x = x
            wall.center_y = 32
            self.wall_list.append(wall)

        # This is an example of how to load an InputManager from a file. The parse function
        # accepts a dictionary, so anything such as toml, yaml, or json could be used to load
        # that dictionary. As long as it can handle a Python dictionary.
        #
        # with open("out.json", "r") as f:
        #     raw = json.load(f)
        # self.INPUT_TEMPLATE = arcade.InputManager.parse(raw)
        # self.INPUT_TEMPLATE.allow_keyboard = False

        self.INPUT_TEMPLATE = InputManager(allow_keyboard=False)
        self.INPUT_TEMPLATE.new_action("Jump")
        self.INPUT_TEMPLATE.add_action_input("Jump", Keys.SPACE)
        self.INPUT_TEMPLATE.add_action_input("Jump", ControllerButtons.BOTTOM_FACE)

        self.INPUT_TEMPLATE.new_axis("Move")
        self.INPUT_TEMPLATE.add_axis_input("Move", Keys.A, -1.0)
        self.INPUT_TEMPLATE.add_axis_input("Move", Keys.D, 1.0)
        self.INPUT_TEMPLATE.add_axis_input("Move", ControllerAxes.LEFT_STICK_X)

        # This is an example of how to dump an InputManager to a file. The serialize function
        # returns a dictionary, so anything such as toml, yaml, or json could be used to save
        # that dictionary. As long as it can handle a Python dictionary.
        #
        # serialized = self.INPUT_TEMPLATE.serialize()
        # with open("out.json", "w") as f:
        #     json.dump(serialized, f)

        self.controller_manager = pyglet.input.ControllerManager()
        self.controller_manager.set_handlers(self.on_connect, self.on_disconnect)

        for index, controller in enumerate(self.controller_manager.get_controllers()):
            self.add_player_for_controller(controller)

    def _get_first_player_slot_free(self) -> int | None:
        for i, player in self.players:
            if player is None:
                return i

        if len(self.players) < self._max_players:
            return len(self.players)

        return None

    def add_player_for_controller(self, controller: Controller) -> None:
        slot = self._get_first_player_slot_free()
        texture = self.player_textures[slot]
        print(f"got {slot=}, {texture=}")

        # Put the player in a random location
        player = Player(
            texture,
            center_x=random.randrange(0, self.width),
            center_y=128,
            walls=self.wall_list,
            input_manager_template=self.INPUT_TEMPLATE,
            controller=controller,
        )

        label_text = f"Player {slot + 1}\n({controller.device.name})"
        label = arcade.Text(
            label_text,
            player.center_x,
            player.center_y + FLOAT_HEIGHT,
            multiline=True,
            width=400,
            align="center",
            anchor_x="center",
            batch=self.device_labels_batch,
        )

        if slot == len(self.players):
            self.players.append(player)
            self.player_device_labels.append(label)
        else:
            self.players[slot] = player
            self.player_device_labels[slot] = label

        self.player_list.append(player)

    def on_connect(self, controller: pyglet.input.Controller):
        self.add_player_for_controller(controller)

    def on_disconnect(self, controller: pyglet.input.Controller):
        to_remove = None
        to_remove_index = -1

        for index, player in enumerate(self.players):
            if player.input_manager.controller == controller:
                to_remove = player
                to_remove_index = index
                break

        if to_remove is None:
            return

        to_remove.input_manager.unbind_controller()
        to_remove.kill()
        self.players[to_remove_index] = None
        # self.device_labels_batch.remove()
        self.player_device_labels[to_remove_index] = None

    def on_draw(self):
        self.clear()

        self.player_list.draw()
        self.wall_list.draw()
        self.device_labels_batch.draw()

    def on_key_press(self, key, modifiers):
        player_index = self.key_to_player_index.get(Keys(key), None)
        if player_index is None or player_index >= len(self.players):
            return

        # Give keyboard focus to player # N
        self.players[player_index].input_manager.allow_keyboard = True
        for index, player in enumerate(self.players):
            if index != player_index:
                player.input_manager.allow_keyboard = False

    def on_update(self, delta_time: float):
        self.player_list.update(delta_time)
        for label, player in zip(self.player_device_labels, self.players):
            position_x, position_y = player.position
            position_y += FLOAT_HEIGHT
            label.position = position_x, position_y


def main():
    Game()
    arcade.run()


if __name__ == "__main__":
    main()



"""
Performance Statistic Display Example

This example demonstrates how to use a few performance profiling tools
built into Arcade:

* arcade.enable_timings
* arcade.PerfGraph
* arcade.get_fps
* arcade.print_timings
* arcade.clear_timings

A large number of sprites bounce around the screen to produce load. You
can adjust the number of sprites by changing the COIN_COUNT constant.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the
command line with:
python -m arcade.examples.performance_statistics
"""
import random

import arcade

# --- Constants ---
SPRITE_SCALING_COIN = 0.25
SPRITE_NATIVE_SIZE = 128
SPRITE_SIZE = int(SPRITE_NATIVE_SIZE * SPRITE_SCALING_COIN)

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Performance Statistics Display Example"

# Size of performance graphs and distance between them
GRAPH_WIDTH = 200
GRAPH_HEIGHT = 120
GRAPH_MARGIN = 5

COIN_COUNT = 1500

# Turn on tracking for the number of event handler
# calls and the average execution time of each type.
arcade.enable_timings()


class Coin(arcade.BasicSprite):
    """ Our coin sprite class """
    def __init__(self, texture: arcade.Texture, scale: float):
        super().__init__(texture, scale=scale)
        # Add a velocity to the coin
        self.change_x = 0
        self.change_y = 0

    def update(self, delta_time: float = 1/60):
        """ Update the sprite. """
        # Setting the position is faster than setting x & y individually
        self.position = (
            self.position[0] + self.change_x,
            self.position[1] + self.change_y
        )

        # Bounce the coin on the edge of the window
        if self.position[0] < 0:
            self.change_x *= -1
        elif self.position[0] > WINDOW_WIDTH:
            self.change_x *= -1
        if self.position[1] < 0:
            self.change_y *= -1
        elif self.position[1] > WINDOW_HEIGHT:
            self.change_y *= -1


class GameView(arcade.View):
    """ Our custom Window Class"""

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables to hold game objects and performance info
        self.coin_list: arcade.SpriteList | None = None
        self.perf_graph_list: arcade.SpriteList | None = None
        self.fps_text: arcade.Text | None = None
        self.frame_count: int = 0  # for tracking the reset interval

        self.coin_texture = arcade.load_texture(":resources:images/items/coinGold.png")
        self.background_color = arcade.color.AMAZON

    def add_coins(self, amount):

        # Create the coins
        for i in range(amount):
            # Create the coin instance
            # Coin image from kenney.nl
            coin = Coin(self.coin_texture, scale=SPRITE_SCALING_COIN)

            # Position the coin
            coin.position = (
                random.randrange(SPRITE_SIZE, WINDOW_WIDTH - SPRITE_SIZE),
                random.randrange(SPRITE_SIZE, WINDOW_HEIGHT - SPRITE_SIZE)
            )

            coin.change_x = random.randrange(-3, 4)
            coin.change_y = random.randrange(-3, 4)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.coin_list = arcade.SpriteList(use_spatial_hash=False)

        # Create some coins
        self.add_coins(COIN_COUNT)

        # Create a sprite list to put the performance graphs into
        self.perf_graph_list = arcade.SpriteList()

        # Calculate position helpers for the row of 3 performance graphs
        row_y = self.height - GRAPH_HEIGHT / 2
        starting_x = GRAPH_WIDTH / 2
        step_x = GRAPH_WIDTH + GRAPH_MARGIN

        # Create the FPS performance graph
        graph = arcade.PerfGraph(GRAPH_WIDTH, GRAPH_HEIGHT, graph_data="FPS")
        graph.position = starting_x, row_y
        self.perf_graph_list.append(graph)

        # Create the on_update graph
        graph = arcade.PerfGraph(GRAPH_WIDTH, GRAPH_HEIGHT, graph_data="on_update")
        graph.position = starting_x + step_x, row_y
        self.perf_graph_list.append(graph)

        # Create the on_draw graph
        graph = arcade.PerfGraph(GRAPH_WIDTH, GRAPH_HEIGHT, graph_data="on_draw")
        graph.position = starting_x + step_x * 2, row_y
        self.perf_graph_list.append(graph)

        # Create a Text object to show the current FPS
        self.fps_text = arcade.Text(
            f"FPS: {arcade.get_fps(60):5.1f}",
            10, 10, arcade.color.BLACK, 22
        )

    def on_draw(self):
        """ Draw everything """

        # Clear the screen
        self.clear()

        # Draw all the coin sprites
        self.coin_list.draw()

        # Draw the graphs
        self.perf_graph_list.draw()

        # Get & draw the FPS for the last 60 frames
        if arcade.timings_enabled():
            self.fps_text.value = f"FPS: {arcade.get_fps(60):5.1f}"
            self.fps_text.draw()

    def on_update(self, delta_time):
        """ Update method """
        self.frame_count += 1

        # Print and clear timings every 60 frames
        if self.frame_count % 60 == 0:
            arcade.print_timings()
            arcade.clear_timings()

        # Move the coins
        self.coin_list.update()

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.SPACE:
            if arcade.timings_enabled():
                arcade.disable_timings()
            else:
                arcade.enable_timings()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()

if __name__ == "__main__":
    main()



"""
"Hook" file for pyinstaller - enables Arcade to be easily packaged with PyInstaller.

Docs: https://pyinstaller.readthedocs.io/en/stable/hooks.html
Official Sample: https://github.com/pyinstaller/hooksample

Note: See setup.cfg for entry points that enable this hook file to be found

For a walk through of building an Arcade application with pyinstaller see:
https://api.arcade.academy/en/latest/tutorials/bundling_with_pyinstaller/index.html
"""

from importlib.util import find_spec
from pathlib import Path

from PyInstaller.compat import is_darwin, is_unix, is_win

# check for supported operating systems
if not is_win and not is_darwin and not is_unix:
    raise NotImplementedError(
        "You are running on an unsupported operating system. "
        "Only Linux, Mac, and Windows are supported."
    )

datas = []
binaries = []

# Add Arcade resources
arcade_spec = find_spec("arcade")
if arcade_spec is None or arcade_spec.origin is None:
    raise ImportError("Arcade is not installed. Cannot continue.")

arcade_path = Path(arcade_spec.origin).parent
datas.extend(
    [
        (
            arcade_path / "resources" / "system",
            "./arcade/resources/system",
        ),
        (
            arcade_path / "VERSION",
            "./arcade/VERSION",
        ),
    ]
)

if is_darwin:
    binaries.append((arcade_path / "lib", "./arcade/lib"))

# Add Pymunk resources
pymunk_spec = find_spec("pymunk")
if pymunk_spec is not None and pymunk_spec.origin is not None:
    pymunk_path = Path(pymunk_spec.origin).parent

    if is_win:
        binaries.append((pymunk_path / "_chipmunk.pyd", "."))
    elif is_darwin:
        binaries.append((pymunk_path / "_chipmunk.abi3.so", "."))
    elif is_unix:
        binaries.append((pymunk_path / "_chipmunk.abi3.so", "."))
else:
    print("Pymunk is not available. Skipping Pymunk resources.")



"""
Sound Speed Demo

If Python and Arcade are installed, this example can be run from the
command line with:
python -m arcade.examples.sound_speed_demo

Left click a button to play a sound.

Each button plays the same sound sample in a slightly different way.

The middle buttons play at normal speed. The ones to the left play
slower, and the ones to the right play faster. The buttons higher on
the screen are quieter, while the ones further down are louder.
"""

import typing

import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sound Speed Demo"
BUTTON_SIZE = 30


SPEED_VARIATION = [0.1, 0.5, 1.0, 2.0, 4.0]
MARGIN = WINDOW_WIDTH / 4
BUTTON_X_POSITIONS = [
    MARGIN,
    MARGIN + (WINDOW_WIDTH - MARGIN * 2) / 3 * 1,
    MARGIN + (WINDOW_WIDTH - MARGIN * 2) / 3 * 2,
    MARGIN + (WINDOW_WIDTH - MARGIN * 2) / 3 * 3,
    WINDOW_WIDTH - MARGIN,
]


VOLUME_VARIATION = [0.1, 0.5, 1]
Y_OFFSETS = [50, 0, -50]


class SoundButton(arcade.SpriteSolidColor):
    """
    A sprite that stores settings about how to play a sound.

    You can tell it to play a sound faster or slower, as well as adjust
    the volume of the sound.
    """
    def __init__(self, sound_file, speed, volume, center_x, center_y):
        super().__init__(BUTTON_SIZE, BUTTON_SIZE, color=arcade.color.WHITE)
        self.sound = arcade.Sound(sound_file)
        self.speed = speed
        self.volume = volume
        self.center_x = center_x
        self.center_y = center_y

    def play(self):
        self.sound.play(speed=self.speed, volume=self.volume)


class GameView(arcade.View):
    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.AMAZON
        self.button_sprites = arcade.SpriteList()

        # Position the grid of buttons
        # The zip function takes pieces from iterables and returns them
        # as tuples. For more information, you can see the python doc:
        # https://docs.python.org/3/library/functions.html#zip
        for vol, y_offset in zip(VOLUME_VARIATION, Y_OFFSETS):
            for speed, x_pos in zip(SPEED_VARIATION, BUTTON_X_POSITIONS):
                self.button_sprites.append(
                    SoundButton(
                        ":resources:sounds/gameover3.wav",
                        speed,
                        vol,
                        x_pos,
                        WINDOW_HEIGHT / 2 + y_offset,
                    )
                )

    def on_draw(self):
        self.clear()
        self.button_sprites.draw()

    def on_update(self, delta_time):
        self.button_sprites.update()

    def on_mouse_press(self, x, y, button, key_modifiers):
        hit_sprites = arcade.get_sprites_at_point((x, y), self.button_sprites)
        for sprite in hit_sprites:
            button_sprite = typing.cast(SoundButton, sprite)
            if button == arcade.MOUSE_BUTTON_LEFT:
                button_sprite.play()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Experimental video player using pyglet.

This requires that you have ffmpeg installed
and you might need to tell pyglet where it's located.
"""

from pathlib import Path

# import sys
import pyglet

import arcade


class VideoPlayer:
    """
    Primitive video player for arcade.

    Args:
        path: Path of the video that is to be played.
        loop: Pass `True` to make the video loop.
    """

    def __init__(self, path: str | Path, loop: bool = False):
        self.player = pyglet.media.Player()
        self.player.loop = loop
        self.player.queue(pyglet.media.load(str(arcade.resources.resolve(path))))
        self.player.play()

        self.ctx = arcade.get_window().ctx

        self._width = arcade.get_window().width
        self._height = arcade.get_window().height

    def draw(self, left: int = 0, bottom: int = 0, size: tuple[int, int] | None = None) -> None:
        """
        Draw the current video frame.

        Args:
            left:
                Window position from the left.
            bottom:
                Window position from the bottom.
            size:
                The size of the video rectangle.
                If `None`, the video will be drawn in its original size.
        """
        if size and len(size) == 2:
            self._width = size[0] or self.width
            self._height = size[1] or self.height

        self.ctx.disable(self.ctx.BLEND)
        video_texture = self.player.texture
        if video_texture:
            video_texture.blit(
                left,
                bottom,
                width=self.width,
                height=self.height,
            )

    @property
    def width(self) -> int:
        """Video width."""
        return self._width

    @property
    def height(self) -> int:
        """Video height."""
        return self._height

    def get_video_size(self) -> tuple[int, int]:
        if not self.player.source or not self.player.source.video_format:
            return 0, 0
        video_format = self.player.source.video_format
        width = video_format.width
        height = video_format.height
        if video_format.sample_aspect > 1:
            width *= video_format.sample_aspect
        elif video_format.sample_aspect < 1:
            height /= video_format.sample_aspect
        return width, height


class VideoPlayerView(arcade.View):
    def __init__(self, path: str | Path) -> None:
        super().__init__()
        self.video_player = VideoPlayer(path)

    def on_draw(self) -> None:
        self.clear()
        self.video_player.draw()


if __name__ == "__main__":
    window = arcade.Window(1280, 720, "Video Player")
    window.show_view(VideoPlayerView(":resources:video/earth.mp4"))
    window.run()



"""
Drawing an example happy face

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.happy_face
"""

import arcade

# Set constants for the screen size
WINDOW_WIDTH = 600
WINDOW_HEIGHT = 600
WINDOW_TITLE = "Happy Face Example"

# Open the window. Set the window title and dimensions
arcade.open_window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

# Set the background color
arcade.set_background_color(arcade.color.WHITE)

# Clear screen and start render process
arcade.start_render()

# --- Drawing Commands Will Go Here ---

# Draw the face
x = 300
y = 300
radius = 200
arcade.draw_circle_filled(x, y, radius, arcade.color.YELLOW)

# Draw the right eye
x = 370
y = 350
radius = 20
arcade.draw_circle_filled(x, y, radius, arcade.color.BLACK)

# Draw the left eye
x = 230
y = 350
radius = 20
arcade.draw_circle_filled(x, y, radius, arcade.color.BLACK)

# Draw the smile
x = 300
y = 280
width = 240
height = 200
start_angle = 190
end_angle = 350
line_width = 20
arcade.draw_arc_outline(x, y, width, height, arcade.color.BLACK,
                        start_angle, end_angle, line_width)

# Finish drawing and display the result
arcade.finish_render()

# Keep the window open until the user hits the 'close' button
arcade.run()



import struct
from ctypes import POINTER, c_double, c_float, c_int, c_uint, cast
from typing import Callable

from pyglet import gl

from .exceptions import ShaderException


class Uniform:
    """
    A Program uniform

    Args:
        ctx:
            The context
        program_id:
            The program id to which this uniform belongs
        location:
            The uniform location
        name:
            The uniform name
        data_type:
            The data type of the uniform
        array_length:
            The array length of the uniform
    """

    _type_to_struct = {
        c_float: "f",
        c_int: "i",
        c_uint: "I",
        c_double: "d",
    }

    _uniform_getters = {
        gl.GLint: gl.glGetUniformiv,
        gl.GLuint: gl.glGetUniformuiv,
        gl.GLfloat: gl.glGetUniformfv,
    }

    _uniform_setters = {
        # uniform type: (gl_type, setter, length, count)
        # Integers 32 bit
        gl.GL_INT: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_INT_VEC2: (gl.GLint, gl.glProgramUniform2iv, gl.glUniform2iv, 2, 1),
        gl.GL_INT_VEC3: (gl.GLint, gl.glProgramUniform3iv, gl.glUniform3iv, 3, 1),
        gl.GL_INT_VEC4: (gl.GLint, gl.glProgramUniform4iv, gl.glUniform4iv, 4, 1),
        # Unsigned integers 32 bit
        gl.GL_UNSIGNED_INT: (gl.GLuint, gl.glProgramUniform1uiv, gl.glUniform1uiv, 1, 1),
        gl.GL_UNSIGNED_INT_VEC2: (gl.GLuint, gl.glProgramUniform2uiv, gl.glUniform2uiv, 2, 1),
        gl.GL_UNSIGNED_INT_VEC3: (gl.GLuint, gl.glProgramUniform3uiv, gl.glUniform3uiv, 3, 1),
        gl.GL_UNSIGNED_INT_VEC4: (gl.GLuint, gl.glProgramUniform4uiv, gl.glUniform4uiv, 4, 1),
        # Integers 64 bit unsigned
        gl.GL_INT64_ARB: (gl.GLint64, gl.glProgramUniform1i64vARB, gl.glUniform1i64vARB, 1, 1),
        gl.GL_INT64_VEC2_ARB: (gl.GLint64, gl.glProgramUniform2i64vARB, gl.glUniform2i64vARB, 2, 1),
        gl.GL_INT64_VEC3_ARB: (gl.GLint64, gl.glProgramUniform3i64vARB, gl.glUniform3i64vARB, 3, 1),
        gl.GL_INT64_VEC4_ARB: (gl.GLint64, gl.glProgramUniform4i64vARB, gl.glUniform4i64vARB, 4, 1),
        # Unsigned integers 64 bit
        gl.GL_UNSIGNED_INT64_ARB: (
            gl.GLuint64,
            gl.glProgramUniform1ui64vARB,
            gl.glUniform1ui64vARB,
            1,
            1,
        ),
        gl.GL_UNSIGNED_INT64_VEC2_ARB: (
            gl.GLuint64,
            gl.glProgramUniform2ui64vARB,
            gl.glUniform2ui64vARB,
            2,
            1,
        ),
        gl.GL_UNSIGNED_INT64_VEC3_ARB: (
            gl.GLuint64,
            gl.glProgramUniform3ui64vARB,
            gl.glUniform3ui64vARB,
            3,
            1,
        ),
        gl.GL_UNSIGNED_INT64_VEC4_ARB: (
            gl.GLuint64,
            gl.glProgramUniform4ui64vARB,
            gl.glUniform4ui64vARB,
            4,
            1,
        ),
        # Bools
        gl.GL_BOOL: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_BOOL_VEC2: (gl.GLint, gl.glProgramUniform2iv, gl.glUniform2iv, 2, 1),
        gl.GL_BOOL_VEC3: (gl.GLint, gl.glProgramUniform3iv, gl.glUniform3iv, 3, 1),
        gl.GL_BOOL_VEC4: (gl.GLint, gl.glProgramUniform4iv, gl.glUniform4iv, 4, 1),
        # Floats 32 bit
        gl.GL_FLOAT: (gl.GLfloat, gl.glProgramUniform1fv, gl.glUniform1fv, 1, 1),
        gl.GL_FLOAT_VEC2: (gl.GLfloat, gl.glProgramUniform2fv, gl.glUniform2fv, 2, 1),
        gl.GL_FLOAT_VEC3: (gl.GLfloat, gl.glProgramUniform3fv, gl.glUniform3fv, 3, 1),
        gl.GL_FLOAT_VEC4: (gl.GLfloat, gl.glProgramUniform4fv, gl.glUniform4fv, 4, 1),
        # Floats 64 bit
        gl.GL_DOUBLE: (gl.GLdouble, gl.glProgramUniform1dv, gl.glUniform1dv, 1, 1),
        gl.GL_DOUBLE_VEC2: (gl.GLdouble, gl.glProgramUniform2dv, gl.glUniform2dv, 2, 1),
        gl.GL_DOUBLE_VEC3: (gl.GLdouble, gl.glProgramUniform3dv, gl.glUniform3dv, 3, 1),
        gl.GL_DOUBLE_VEC4: (gl.GLdouble, gl.glProgramUniform4dv, gl.glUniform4dv, 4, 1),
        # 1D Samplers
        gl.GL_SAMPLER_1D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_INT_SAMPLER_1D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_UNSIGNED_INT_SAMPLER_1D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_TEXTURE_1D_ARRAY: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        # 2D samplers
        gl.GL_SAMPLER_2D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_SAMPLER_2D_MULTISAMPLE: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_INT_SAMPLER_2D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_UNSIGNED_INT_SAMPLER_2D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_TEXTURE_2D_MULTISAMPLE: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        # Array
        gl.GL_SAMPLER_2D_ARRAY: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_TEXTURE_2D_MULTISAMPLE_ARRAY: (
            gl.GLint,
            gl.glProgramUniform1iv,
            gl.glUniform1iv,
            1,
            1,
        ),
        # 3D
        gl.GL_SAMPLER_3D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        # Cube
        gl.GL_SAMPLER_CUBE: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_TEXTURE_CUBE_MAP_ARRAY: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        # Matrices
        gl.GL_FLOAT_MAT2: (gl.GLfloat, gl.glProgramUniformMatrix2fv, gl.glUniformMatrix2fv, 4, 1),
        gl.GL_FLOAT_MAT3: (gl.GLfloat, gl.glProgramUniformMatrix3fv, gl.glUniformMatrix3fv, 9, 1),
        gl.GL_FLOAT_MAT4: (gl.GLfloat, gl.glProgramUniformMatrix4fv, gl.glUniformMatrix4fv, 16, 1),
        # Image (compute shader)
        gl.GL_IMAGE_1D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_2D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_2D_RECT: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_3D: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_CUBE: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_1D_ARRAY: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_2D_ARRAY: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_CUBE_MAP_ARRAY: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_2D_MULTISAMPLE: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_2D_MULTISAMPLE_ARRAY: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        gl.GL_IMAGE_BUFFER: (gl.GLint, gl.glProgramUniform1iv, gl.glUniform1iv, 1, 1),
        # TODO: test/implement these:
        # gl.GL_FLOAT_MAT2x3: glUniformMatrix2x3fv,
        # gl.GL_FLOAT_MAT2x4: glUniformMatrix2x4fv,
        #
        # gl.GL_FLOAT_MAT3x2: glUniformMatrix3x2fv,
        # gl.GL_FLOAT_MAT3x4: glUniformMatrix3x4fv,
        #
        # gl.GL_FLOAT_MAT4x2: glUniformMatrix4x2fv,
        # gl.GL_FLOAT_MAT4x3: glUniformMatrix4x3fv,
    }

    __slots__ = (
        "_program_id",
        "_location",
        "_name",
        "_data_type",
        "_array_length",
        "_components",
        "getter",
        "setter",
        "_ctx",
    )

    def __init__(self, ctx, program_id, location, name, data_type, array_length):
        self._ctx = ctx
        self._program_id = program_id
        self._location = location
        self._name = name
        self._data_type = data_type
        # Array length of the uniform (1 if no array)
        self._array_length = array_length
        # Number of components (including per array entry)
        self._components = 0
        self.getter: Callable
        """The getter function configured for this uniform"""
        self.setter: Callable
        """The setter function configured for this uniform"""
        self._setup_getters_and_setters()

    @property
    def location(self) -> int:
        """The location of the uniform in the program"""
        return self._location

    @property
    def name(self) -> str:
        """Name of the uniform"""
        return self._name

    @property
    def array_length(self) -> int:
        """Length of the uniform array. If not an array 1 will be returned"""
        return self._array_length

    @property
    def components(self) -> int:
        """
        How many components for the uniform.

        A vec4 will for example have 4 components.
        """
        return self._components

    def _setup_getters_and_setters(self):
        """Maps the right getter and setter functions for this uniform"""
        try:
            gl_type, gl_program_setter, gl_setter, length, count = self._uniform_setters[
                self._data_type
            ]
            self._components = length
        except KeyError:
            raise ShaderException(f"Unsupported Uniform type: {self._data_type}")

        gl_getter = self._uniform_getters[gl_type]
        is_matrix = self._data_type in (
            gl.GL_FLOAT_MAT2,
            gl.GL_FLOAT_MAT3,
            gl.GL_FLOAT_MAT4,
        )

        # Create persistent mini c_array for getters and setters:
        length = length * self._array_length  # Increase buffer size to include arrays
        c_array = (gl_type * length)()
        ptr = cast(c_array, POINTER(gl_type))

        # Create custom dedicated getters and setters for each uniform:
        self.getter = Uniform._create_getter_func(
            self._program_id,
            self._location,
            gl_getter,
            c_array,
            length,
        )

        self.setter = Uniform._create_setter_func(
            self._ctx,
            self._program_id,
            self._location,
            gl_program_setter,
            gl_setter,
            c_array,
            gl_type,
            length,
            self._array_length,
            count,
            ptr,
            is_matrix,
        )

    @staticmethod
    def _create_getter_func(program_id, location, gl_getter, c_array, length):
        """Create a function for getting/setting OpenGL data."""

        def getter_func1():
            """Get single-element OpenGL uniform data."""
            gl_getter(program_id, location, c_array)
            return c_array[0]

        def getter_func2():
            """Get list of OpenGL uniform data."""
            gl_getter(program_id, location, c_array)
            return tuple(c_array)

        if length == 1:
            return getter_func1
        else:
            return getter_func2

    @classmethod
    def _create_setter_func(
        cls,
        ctx,
        program_id,
        location,
        gl_program_setter,
        gl_setter,
        c_array,
        gl_type,
        length,
        array_length,
        count,
        ptr,
        is_matrix,
    ):
        """Create setters for OpenGL data."""
        # Matrix uniforms
        if is_matrix:
            if ctx._ext_separate_shader_objects_enabled:

                def setter_func(value):  # type: ignore #conditional function variants must have identical signature
                    """Set OpenGL matrix uniform data."""
                    try:
                        # FIXME: Configure the struct format on the uniform to support
                        #        other types than float
                        fmt = cls._type_to_struct[gl_type]
                        c_array[:] = struct.unpack(f"{length}{fmt}", value)
                    except Exception:
                        c_array[:] = value
                    gl_program_setter(program_id, location, array_length, gl.GL_FALSE, ptr)

            else:

                def setter_func(value):  # type: ignore #conditional function variants must have identical signature
                    """Set OpenGL matrix uniform data."""
                    c_array[:] = value
                    gl.glUseProgram(program_id)
                    gl_setter(location, array_length, gl.GL_FALSE, ptr)

        # Single value uniforms
        elif length == 1 and count == 1:
            if ctx._ext_separate_shader_objects_enabled:

                def setter_func(value):  # type: ignore #conditional function variants must have identical signature
                    """Set OpenGL uniform data value."""
                    c_array[0] = value
                    gl_program_setter(program_id, location, array_length, ptr)

            else:

                def setter_func(value):  # type: ignore #conditional function variants must have identical signature
                    """Set OpenGL uniform data value."""
                    c_array[0] = value
                    gl.glUseProgram(program_id)
                    gl_setter(location, array_length, ptr)

        # Uniforms types with multiple components
        elif length > 1 and count == 1:
            if ctx._ext_separate_shader_objects_enabled:

                def setter_func(values):  # type: ignore #conditional function variants must have identical signature
                    """Set list of OpenGL uniform data."""
                    # Support buffer protocol
                    try:
                        # FIXME: Configure the struct format on the uniform to support
                        #        other types than float
                        fmt = cls._type_to_struct[gl_type]
                        c_array[:] = struct.unpack(f"{length}{fmt}", values)
                    except Exception:
                        c_array[:] = values

                    gl_program_setter(program_id, location, array_length, ptr)

            else:

                def setter_func(values):  # type: ignore #conditional function variants must have identical signature
                    """Set list of OpenGL uniform data."""
                    c_array[:] = values
                    gl.glUseProgram(program_id)
                    gl_setter(location, array_length, ptr)

        else:
            raise NotImplementedError("Uniform type not yet supported.")

        return setter_func

    def __repr__(self) -> str:
        return f"<Uniform '{self._name}' loc={self._location} array_length={self._array_length}>"


class UniformBlock:
    """
    Wrapper for a uniform block in shaders.

    Args:
        glo:
            The OpenGL object handle
        index:
            The index of the uniform block
        size:
            The size of the uniform block
        name:
            The name of the uniform
    """

    __slots__ = ("glo", "index", "size", "name")

    def __init__(self, glo: int, index: int, size: int, name: str):
        self.glo = glo
        """The OpenGL object handle"""

        self.index = index
        """The index of the uniform block"""

        self.size = size
        """The size of the uniform block"""

        self.name = name
        """The name of the uniform block"""

    @property
    def binding(self) -> int:
        """Get or set the binding index for this uniform block"""
        binding = gl.GLint()
        gl.glGetActiveUniformBlockiv(self.glo, self.index, gl.GL_UNIFORM_BLOCK_BINDING, binding)
        return binding.value

    @binding.setter
    def binding(self, binding: int):
        gl.glUniformBlockBinding(self.glo, self.index, binding)

    def getter(self):
        """
        The getter function for this uniform block.

        Returns self.
        """
        return self

    def setter(self, value: int):
        """
        The setter function for this uniform block.

        Args:
            value: The binding index to set.
        """
        self.binding = value

    def __str__(self) -> str:
        return f"<UniformBlock {self.name} index={self.index} size={self.size}>"



"""
This module provides a Scene class which acts as a sprite list manager.

It allows you to do the following:

* Group sprite lists into a larger object, using them like layers
* Draw & update the contained sprite lists
* Add sprites by name
* Load from tiled map objects
* Control sprite list draw order within the group
"""

from collections.abc import Iterable
from warnings import warn

from arcade import Sprite, SpriteList
from arcade.gl.types import BlendFunction, OpenGlFilter
from arcade.tilemap import TileMap
from arcade.types import Color, RGBOrA255

__all__ = ["Scene", "SceneKeyError"]


class SceneKeyError(KeyError):
    """
    Raised when a py:class:`.Scene` cannot find a layer for a specified name.

    It is a subclass of :py:class:`KeyError`, and you can handle it as
    one if you wish::

        try:
            # this will raise a SceneKeyError
            scene_instance.add_sprite("missing_layer_name", arcade.SpriteSolidColor(10,10))

        # We can handle it as a KeyError because it is a subclass of it
        except KeyError as e:
            print("Your error handling should go here")

    The main purpose of this class is to help Arcade's developers keep
    error messages consistent.

    Args:
        name: the name of the missing :py:class:`~arcade.SpriteList`
    """

    def __init__(self, name: str):
        super().__init__(f"This scene does not contain a SpriteList named {name!r}.")
        self.layer_name = name


class Scene:
    """
    Stores :py:class:`~arcade.SpriteList` instances as named layers,
    allowing bulk updates & drawing.

    In addition to helping you update or draw multiple sprite lists at
    once, this class also provides the following convenience methods:

    * :py:meth:`.add_sprite`, which adds sprites to layers by name
    * :py:meth:`.Scene.from_tilemap`, which creates a scene from a
      :py:class:`~arcade.tilemap.TileMap` already loaded from tiled data
    * Fine-grained convenience methods for adding, deleting, and reordering
      sprite lists
    * Flexible but slow general convenience methods
    * Flexible but slow support for the ``in`` & ``del`` Python keywords.

    For another example of how to use this class, see :ref:`platformer_part_three`.
    """

    def __init__(self) -> None:
        self._sprite_lists: list[SpriteList] = []
        self._name_mapping: dict[str, SpriteList] = {}

    def __len__(self) -> int:
        """
        Return the number of sprite lists in this scene.
        """
        return len(self._sprite_lists)

    def __delitem__(self, sprite_list: int | str | SpriteList) -> None:
        """
        Remove a sprite list from this scene by its index, name, or instance value.

        .. tip:: Use a more specific method when speed is important!

                 This method uses :py:func:`isinstance`, which will
                 slow down your program if used frequently!

        Consider the following alternatives:

        * :py:meth:`.remove_sprite_list_by_index`
        * :py:meth:`.remove_sprite_list_by_name`
        * :py:meth:`.remove_sprite_list_by_object`

        Args:
            sprite_list:
                The index, name, or :py:class:`~arcade.SpriteList` instance to remove from
                this scene.
        """
        if isinstance(sprite_list, int):
            self.remove_sprite_list_by_index(sprite_list)
        elif isinstance(sprite_list, str):
            self.remove_sprite_list_by_name(sprite_list)
        else:
            self.remove_sprite_list_by_object(sprite_list)

    @classmethod
    def from_tilemap(cls, tilemap: TileMap) -> "Scene":
        """
        Create a new Scene from a :py:class:`~arcade.tilemap.TileMap` object.

        The SpriteLists will use the layer names and ordering as defined in the
        Tiled file.

        Args:
            tilemap: The :py:class:`~arcade.tilemap.TileMap`
                object to create the scene from.
        """
        scene = cls()
        for name, sprite_list in tilemap.sprite_lists.items():
            scene.add_sprite_list(name=name, sprite_list=sprite_list)
        return scene

    def get_sprite_list(self, name: str) -> SpriteList:
        """
        Retrieve a sprite list by name.

        It is also possible to access sprite lists the following ways:

        * ``scene_instance[name]``
        * directly accessing ``scene_instance._name_mapping``, although this will
          get flagged by linters as bad style.

        Args:
            name: The name of the sprite list to retrieve.
        """
        return self._name_mapping[name]

    def __getitem__(self, key: str) -> SpriteList:
        """
        Retrieve a sprite list by name.

        This is here for ease of use to make sub-scripting the scene object directly
        to retrieve a SpriteList possible.

        Args:
            key: The name of the sprite list to retrieve
        """
        if key in self._name_mapping:
            return self._name_mapping[key]

        raise SceneKeyError(key)

    def add_sprite(self, name: str, sprite: Sprite) -> None:
        """
        Add a Sprite to the SpriteList with the specified name.

        If there is no SpriteList for the given ``name``, one will be
        created with :py:class:`SpriteList`'s default arguments and
        added to the end (top) of the scene's current draw order.

        To fully customize the SpriteList's options, you should create
        it directly and add it to the scene with one of the following:

        * :py:meth:`.add_sprite_list_before`
        * :py:meth:`.add_sprite_list`
        * :py:meth:`.add_sprite_list_after`

        Args:
            name: The name of the sprite list to add to or create.
            sprite: The sprite to add.
        """
        if name in self._name_mapping:
            self._name_mapping[name].append(sprite)
        else:
            new_list: SpriteList = SpriteList()
            new_list.append(sprite)
            self.add_sprite_list(name=name, sprite_list=new_list)

    def add_sprite_list(
        self,
        name: str,
        use_spatial_hash: bool = False,
        sprite_list: SpriteList | None = None,
    ) -> None:
        """
        Add a SpriteList to the scene with the specified name.

        This will add a new SpriteList as a layer above the others in the scene.

        If no SpriteList is supplied via the ``sprite_list`` parameter then a new one will be
        created, and the ``use_spatial_hash`` parameter will be respected for that creation.

        Args:
            name: The name to give the new layer.
            use_spatial_hash: If creating a new sprite list, whether
                to enable spatial hashing on it.
            sprite_list: Use a specific sprite list rather than creating a new one.
        """
        if sprite_list is None:
            sprite_list = SpriteList(use_spatial_hash=use_spatial_hash)
        if name in self._name_mapping.keys():
            self.remove_sprite_list_by_name(name)
            warn(
                f"A Spritelist with the name: '{name}', is already in the scene, "
                "will override Spritelist"
            )
        self._name_mapping[name] = sprite_list
        self._sprite_lists.append(sprite_list)

    def add_sprite_list_before(
        self,
        name: str,
        before: str,
        use_spatial_hash: bool = False,
        sprite_list: SpriteList | None = None,
    ) -> None:
        """
        Add a sprite list to the scene with the specified name before another SpriteList.

        If no sprite list is supplied via the ``sprite_list`` parameter, then a new one
        will be created. Aside from the value of ``use_spatial_hash`` passed to this
        method, it will use the default arguments for a new :py:class:`SpriteList`.

        The added sprite list will be drawn under the sprite list named in ``before``.

        Args:
            name: The name to give the new layer.
            before: The name of the layer to place the new one before.
            use_spatial_hash: If creating a new sprite list, selects
                whether to enable spatial hashing.
            sprite_list: If a sprite list is passed via
                this argument, it will be used instead of creating a new one.
        """
        if sprite_list is None:
            sprite_list = SpriteList(use_spatial_hash=use_spatial_hash)
        if name in self._name_mapping.keys():
            self.remove_sprite_list_by_name(name)
            warn(
                f"A Spritelist with the name: '{name}', "
                "is already in the scene, will override Spritelist"
            )
        self._name_mapping[name] = sprite_list
        before_list = self._name_mapping[before]
        index = self._sprite_lists.index(before_list)
        self._sprite_lists.insert(index, sprite_list)

    def move_sprite_list_before(
        self,
        name: str,
        before: str,
    ) -> None:
        """
        Move a named SpriteList in the scene to be before another SpriteList in the scene.

        A :py:class:`~arcade.scene.SceneKeyError` will be raised if either ``name``
        or ``before`` contain a name not currently in the scene. This exception can
        be handled as a :py:class:`KeyError`.

        Args:
            name: The name of the SpriteList to move.
            before: The name of the SpriteList to place it before.
        """
        if name not in self._name_mapping:
            raise SceneKeyError(name)

        if before not in self._name_mapping:
            raise SceneKeyError(before)

        name_list = self._name_mapping[name]
        before_list = self._name_mapping[before]
        new_index = self._sprite_lists.index(before_list)
        old_index = self._sprite_lists.index(name_list)
        self._sprite_lists.insert(new_index, self._sprite_lists.pop(old_index))

    def add_sprite_list_after(
        self,
        name: str,
        after: str,
        use_spatial_hash: bool = False,
        sprite_list: SpriteList | None = None,
    ) -> None:
        """
        Add a SpriteList to the scene with the specified name after a specific SpriteList.

        If no sprite list is supplied via the ``sprite_list`` parameter, then a new one
        will be created. Aside from the value of ``use_spatial_hash`` passed to this
        method, it will use the default arguments for a new :py:class:`SpriteList`.

        The added sprite list will be drawn above the sprite list named in ``after``.

        Args:
            name: The name to give the layer.
            after: The name of the layer to place the new one after.
            use_spatial_hash: If creating a new sprite list, selects
                whether to enable spatial hashing.
            sprite_list: If a sprite list is passed via
                this argument, it will be used instead of creating a new one.
        """
        if sprite_list is None:
            sprite_list = SpriteList(use_spatial_hash=use_spatial_hash)
        if name in self._name_mapping.keys():
            self.remove_sprite_list_by_name(name)
            warn(
                f"A Spritelist with the name: '{name}', "
                f"is already in the scene, will override Spritelist"
            )
        self._name_mapping[name] = sprite_list
        after_list = self._name_mapping[after]
        index = self._sprite_lists.index(after_list) + 1
        self._sprite_lists.insert(index, sprite_list)

    def move_sprite_list_after(
        self,
        name: str,
        after: str,
    ) -> None:
        """
        Move a named SpriteList in the scene to be after another SpriteList in the scene.

        A :py:class:`~arcade.scene.SceneKeyError` will be raised if either ``name``
        or ``after`` contain a name not currently in the scene. This exception can
        be handled as a :py:class:`KeyError`.

        Args:
            name: The name of the SpriteList to move.
            after: The name of the SpriteList to place it after.
        """
        if name not in self._name_mapping:
            raise SceneKeyError(name)

        if after not in self._name_mapping:
            raise SceneKeyError(after)

        name_list = self._name_mapping[name]
        after_list = self._name_mapping[after]
        new_index = self._sprite_lists.index(after_list) + 1
        old_index = self._sprite_lists.index(name_list)
        self._sprite_lists.insert(new_index, self._sprite_lists.pop(old_index))

    def remove_sprite_list_by_index(self, index: int) -> None:
        """
        Remove a layer from the scene by its index in the draw order.

        Args:
            index: The index of the sprite list to remove.
        """
        self.remove_sprite_list_by_object(self._sprite_lists[index])

    def remove_sprite_list_by_name(
        self,
        name: str,
    ) -> None:
        """
        Remove a layer from the scene by its name.

        A :py:class:`KeyError` will be raised if the SpriteList is not
        in the scene.

        Args:
            name: The name of the sprite list to remove.
        """
        sprite_list = self._name_mapping[name]
        self._sprite_lists.remove(sprite_list)
        del self._name_mapping[name]

    def remove_sprite_list_by_object(self, sprite_list: SpriteList) -> None:
        """
        Remove the passed SpriteList instance from the Scene.

        A :py:class:`ValueError` will be raised if the passed sprite
        list is not in the scene.

        Args:
            sprite_list: The sprite list to remove.
        """
        self._sprite_lists.remove(sprite_list)
        self._name_mapping = {
            key: val for key, val in self._name_mapping.items() if val != sprite_list
        }

    def update(
        self,
        delta_time: float,
        names: Iterable[str] | None = None,
        *args,
        **kwargs,
    ) -> None:
        """
        Call :py:meth:`~arcade.SpriteList.update` on the scene's sprite lists.

        By default, this method calls :py:meth:`~arcade.SpriteList.update`
        on the scene's sprite lists in the default draw order.

        You can limit and reorder the updates with the ``names``
        argument by passing a list of names in the scene. The sprite
        lists will be drawn in the order of the passed iterable. If a
        name is not in the scene, a :py:class:`KeyError` will be raised.

        Args:
            delta_time: The time step to update by in seconds.
            names: Which layers & what order to update them in.
            *args: Additional positional arguments propagated down to sprites
            **kwargs: Additional keyword arguments propagated down to sprites
        """
        # Due to api changes in 3.0 we sanity check delta_time
        if not isinstance(delta_time, int | float):
            raise TypeError(
                f"Expected a number for delta_time, but got {type(delta_time)} instead."
            )

        if names is not None:
            # Due to api changes in 3.0 we sanity names
            if not isinstance(names, Iterable):
                raise TypeError(
                    f"Expected an iterable of layer names, but got {type(names)} instead."
                )

            for name in names:
                self._name_mapping[name].update(delta_time, *args, **kwargs)
            return

        for sprite_list in self._sprite_lists:
            sprite_list.update(delta_time, *args, **kwargs)

    def update_animation(
        self, delta_time: float, names: Iterable[str] | None = None, *args, **kwargs
    ) -> None:
        """
        Call :py:meth:`~arcade.SpriteList.update_animation` on the scene's sprite lists.

        By default, this method calls :py:meth:`~arcade.SpriteList.update_animation`
        on each sprite list in the scene in the default draw order.

        You can limit and reorder the updates with the ``names``
        argument by passing a list of names in the scene. The sprite
        lists will be drawn in the order of the passed iterable. If a
        name is not in the scene, a :py:class:`KeyError` will be raised.

        Args:
            delta_time: The time step to update by in seconds.
            names: Which layers & what order to update them in.
            *args: Additional positional arguments propagated down to sprites
            **kwargs: Additional keyword arguments propagated down to sprites
        """
        if names:
            for name in names:
                self._name_mapping[name].update_animation(delta_time, *args, **kwargs)
            return

        for sprite_list in self._sprite_lists:
            sprite_list.update_animation(delta_time, *args, **kwargs)

    def draw(
        self,
        names: Iterable[str] | None = None,
        filter: OpenGlFilter | None = None,
        pixelated: bool = False,
        blend_function: BlendFunction | None = None,
        **kwargs,
    ) -> None:
        """
        Call :py:meth:`~arcade.SpriteList.draw` on the scene's sprite lists.

        By default, this method calls :py:meth:`~arcade.SpriteList.draw`
        on each sprite list in the scene in the default draw order.

        You can limit and reorder the draw calls with the ``names``
        argument by passing a list of names in the scene. The sprite
        lists will be drawn in the order of the passed iterable. If a
        name is not in the scene, a :py:class:`KeyError` will be raised.

        The other named keyword arguments are the same as those of
        :py:meth:`SpriteList.draw() <arcade.SpriteList.draw>`. The
        ``**kwargs`` option is for advanced users who have
        subclassed :py:class:`~arcade.SpriteList`.

        Args:
            names: Which layers to draw & what order to draw them in.
            filter: Optional parameter to set OpenGL filter, such as
                ``gl.GL_NEAREST`` to avoid smoothing.
            pixelated: ``True`` for pixel art and ``False`` for smooth scaling.
            blend_function:
                Use the specified OpenGL blend function while drawing the
                sprite list, such as ``arcade.Window.ctx.BLEND_ADDITIVE``
                or ``arcade.Window.ctx.BLEND_DEFAULT``.
        """
        if names:
            for name in names:
                self._name_mapping[name].draw(
                    filter=filter, pixelated=pixelated, blend_function=blend_function, **kwargs
                )
            return

        for sprite_list in self._sprite_lists:
            sprite_list.draw(
                filter=filter, pixelated=pixelated, blend_function=blend_function, **kwargs
            )

    def draw_hit_boxes(
        self,
        color: RGBOrA255 = Color(0, 0, 0, 255),
        line_thickness: float = 1.0,
        names: Iterable[str] | None = None,
    ) -> None:
        """
        Draw debug hit box outlines for sprites in the scene's layers.

        If ``names`` is a valid iterable of layer names in the scene, then hit boxes
        will be drawn for the specified layers in the order of the passed iterable.

        If `names` is not provided, then every layer's hit boxes will be drawn in the
        order specified.

        Args:
            color: The RGBA color to use to draw the hit boxes with.
            line_thickness: How many pixels thick the hit box outlines should be
            names: Which layers & what order to draw their hit boxes in.
        """
        if names:
            for name in names:
                self._name_mapping[name].draw_hit_boxes(color, line_thickness)
            return

        for sprite_list in self._sprite_lists:
            sprite_list.draw_hit_boxes(color, line_thickness)

    def __bool__(self) -> bool:
        """Returns whether or not `_sprite_lists` contains anything"""
        return bool(self._sprite_lists)

    def __contains__(self, item: str | SpriteList) -> bool:
        """True when `item` is in `_sprite_lists` or is a value in `_name_mapping`"""
        return item in self._sprite_lists or item in self._name_mapping



from arcade import get_window
from arcade.color import TRANSPARENT_BLACK
from arcade.gl import geometry
from arcade.gl.texture import Texture2D
from arcade.types import RGBA255


class RenderTargetTexture:
    """
    An alternative render target to the default window/screen.
    This can be used in creative ways.

    Args:
        width: Width of the render target in pixels
        height: Height of the render target in pixels
    """

    def __init__(self, width: int, height: int):
        self.window = get_window()
        if not self.window:
            raise RuntimeError("No window found")

        self.width = width
        self.height = height
        self._background_color: RGBA255 = TRANSPARENT_BLACK

        self.ctx = self.window.ctx

        self._fbo = self.ctx.framebuffer(
            color_attachments=[self.ctx.texture((width, height), components=4)]
        )

        self._quad_fs = geometry.quad_2d_fs()

    @property
    def texture(self) -> Texture2D:
        """The internal OpenGL texture"""
        return self._fbo.color_attachments[0]

    def clear(self):
        """Clear the texture with the configured background color"""
        self._fbo.clear(color=self._background_color)

    def set_background_color(self, color: RGBA255):
        """Set the background color for the light layer"""
        self._background_color = color

    def resize(self, width: int, height: int):
        """Resize the the internal texture"""
        pixel_scale = self.window.get_pixel_ratio()
        self._size = width * pixel_scale, height * pixel_scale
        self._fbo = self.ctx.framebuffer(
            color_attachments=self.ctx.texture((width, height), components=4)
        )



from __future__ import annotations

import weakref
from ctypes import byref, string_at
from typing import TYPE_CHECKING

from pyglet import gl

from ..types import BufferProtocol
from .buffer import Buffer
from .types import (
    BufferOrBufferProtocol,
    PyGLuint,
    compare_funcs,
    pixel_formats,
    swizzle_enum_to_str,
    swizzle_str_to_enum,
)
from .utils import data_to_ctypes

if TYPE_CHECKING:  # handle import cycle caused by type hinting
    from arcade.gl import Context


class Texture2D:
    """
    An OpenGL 2D texture.
    We can create an empty black texture or a texture from byte data.
    A texture can also be created with different datatypes such as
    float, integer or unsigned integer.

    The best way to create a texture instance is through :py:meth:`arcade.gl.Context.texture`

    Supported ``dtype`` values are::

        # Float formats
        'f1': UNSIGNED_BYTE
        'f2': HALF_FLOAT
        'f4': FLOAT
        # int formats
        'i1': BYTE
        'i2': SHORT
        'i4': INT
        # uint formats
        'u1': UNSIGNED_BYTE
        'u2': UNSIGNED_SHORT
        'u4': UNSIGNED_INT

    Args:
        ctx:
            The context the object belongs to
        size:
            The size of the texture
        components:
            The number of components (1: R, 2: RG, 3: RGB, 4: RGBA)
        dtype:
            The data type of each component: f1, f2, f4 / i1, i2, i4 / u1, u2, u4
        data:
            The texture data. Can be bytes or any object supporting
            the buffer protocol.
        filter:
            The minification/magnification filter of the texture
        wrap_x:
            Wrap mode x
        wrap_y:
            Wrap mode y
        target:
            The texture type (Ignored. Legacy)
        depth:
            creates a depth texture if `True`
        samples:
            Creates a multisampled texture for values > 0.
            This value will be clamped between 0 and the max
            sample capability reported by the drivers.
        immutable:
            Make the storage (not the contents) immutable. This can sometimes be
            required when using textures with compute shaders.
        internal_format:
            The internal format of the texture
        compressed:
            Is the texture compressed?
        compressed_data:
            The raw compressed data
    """

    __slots__ = (
        "_ctx",
        "_glo",
        "_width",
        "_height",
        "_dtype",
        "_target",
        "_components",
        "_alignment",
        "_depth",
        "_compare_func",
        "_format",
        "_internal_format",
        "_type",
        "_component_size",
        "_samples",
        "_filter",
        "_wrap_x",
        "_wrap_y",
        "_anisotropy",
        "_immutable",
        "__weakref__",
        "_compressed",
        "_compressed_data",
    )

    def __init__(
        self,
        ctx: Context,
        size: tuple[int, int],
        *,
        components: int = 4,
        dtype: str = "f1",
        data: BufferProtocol | None = None,
        filter: tuple[PyGLuint, PyGLuint] | None = None,
        wrap_x: PyGLuint | None = None,
        wrap_y: PyGLuint | None = None,
        target=gl.GL_TEXTURE_2D,
        depth=False,
        samples: int = 0,
        immutable: bool = False,
        internal_format: PyGLuint | None = None,
        compressed: bool = False,
        compressed_data: bool = False,
    ):
        self._glo = glo = gl.GLuint()
        self._ctx = ctx
        self._width, self._height = size
        self._dtype = dtype
        self._components = components
        self._component_size = 0
        self._alignment = 1
        self._target = target
        self._samples: int = min(max(0, samples), self._ctx.info.MAX_SAMPLES)
        self._depth: bool = depth
        self._immutable = immutable
        self._compare_func: str | None = None
        self._anisotropy = 1.0
        self._internal_format = internal_format
        self._compressed = compressed
        self._compressed_data = compressed_data
        # Default filters for float and integer textures
        # Integer textures should have NEAREST interpolation
        # by default 3.3 core doesn't really support it consistently.
        if "f" in self._dtype:
            self._filter = gl.GL_LINEAR, gl.GL_LINEAR
        else:
            self._filter = gl.GL_NEAREST, gl.GL_NEAREST
        self._wrap_x = gl.GL_REPEAT
        self._wrap_y = gl.GL_REPEAT

        if self._components not in [1, 2, 3, 4]:
            raise ValueError("Components must be 1, 2, 3 or 4")

        if data and self._samples > 0:
            raise ValueError(
                "Multisampled textures are not writable (cannot be initialized with data)"
            )

        self._target = gl.GL_TEXTURE_2D if self._samples == 0 else gl.GL_TEXTURE_2D_MULTISAMPLE

        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glGenTextures(1, byref(self._glo))

        if self._glo.value == 0:
            raise RuntimeError("Cannot create Texture. OpenGL failed to generate a texture id")

        gl.glBindTexture(self._target, self._glo)

        self._texture_2d(data)

        # Only set texture parameters on non-multisample textures
        if self._samples == 0:
            self.filter = filter or self._filter
            self.wrap_x = wrap_x or self._wrap_x
            self.wrap_y = wrap_y or self._wrap_y

        if self._ctx.gc_mode == "auto":
            weakref.finalize(self, Texture2D.delete_glo, self._ctx, glo)

        self.ctx.stats.incr("texture")

    def resize(self, size: tuple[int, int]):
        """
        Resize the texture. This will re-allocate the internal
        memory and all pixel data will be lost.

        .. note:: Immutable textures cannot be resized.

        Args:
            size: The new size of the texture
        """
        if self._immutable:
            raise ValueError("Immutable textures cannot be resized")

        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)

        self._width, self._height = size

        self._texture_2d(None)

    def __del__(self):
        # Intercept garbage collection if we are using Context.gc()
        if self._ctx.gc_mode == "context_gc" and self._glo.value > 0:
            self._ctx.objects.append(self)

    def _texture_2d(self, data):
        """Create a 2D texture"""
        # Start by resolving the texture format
        try:
            format_info = pixel_formats[self._dtype]
        except KeyError:
            raise ValueError(
                f"dype '{self._dtype}' not support. Supported types are : "
                f"{tuple(pixel_formats.keys())}"
            )
        _format, _internal_format, self._type, self._component_size = format_info
        if data is not None:
            byte_length, data = data_to_ctypes(data)
            self._validate_data_size(data, byte_length, self._width, self._height)

        # If we are dealing with a multisampled texture we have less options
        if self._target == gl.GL_TEXTURE_2D_MULTISAMPLE:
            gl.glTexImage2DMultisample(
                self._target,
                self._samples,
                _internal_format[self._components],
                self._width,
                self._height,
                True,  # Fixed sample locations
            )
            return

        # Make sure we unpack the pixel data with correct alignment
        # or we'll end up with corrupted textures
        gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, self._alignment)
        gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, self._alignment)

        # Create depth 2d texture
        if self._depth:
            gl.glTexImage2D(
                self._target,
                0,  # level
                gl.GL_DEPTH_COMPONENT24,
                self._width,
                self._height,
                0,
                gl.GL_DEPTH_COMPONENT,
                gl.GL_UNSIGNED_INT,  # gl.GL_FLOAT,
                data,
            )
            self.compare_func = "<="
        # Create normal 2d texture
        else:
            try:
                self._format = _format[self._components]
                if self._internal_format is None:
                    self._internal_format = _internal_format[self._components]

                if self._immutable:
                    # Specify immutable storage for this texture.
                    # glTexStorage2D can only be called once
                    gl.glTexStorage2D(
                        self._target,
                        1,  # Levels
                        self._internal_format,
                        self._width,
                        self._height,
                    )
                    if data:
                        self.write(data)
                else:
                    # glTexImage2D can be called multiple times to re-allocate storage
                    # Specify mutable storage for this texture.
                    if self._compressed_data is True:
                        gl.glCompressedTexImage2D(
                            self._target,  # target
                            0,  # level
                            self._internal_format,  # internal_format
                            self._width,  # width
                            self._height,  # height
                            0,  # border
                            len(data),  # size
                            data,  # data
                        )
                    else:
                        gl.glTexImage2D(
                            self._target,  # target
                            0,  # level
                            self._internal_format,  # internal_format
                            self._width,  # width
                            self._height,  # height
                            0,  # border
                            self._format,  # format
                            self._type,  # type
                            data,  # data
                        )
            except gl.GLException as ex:
                raise gl.GLException(
                    (
                        f"Unable to create texture: {ex} : dtype={self._dtype} "
                        f"size={self.size} components={self._components} "
                        f"MAX_TEXTURE_SIZE = {self.ctx.info.MAX_TEXTURE_SIZE}"
                        f": {ex}"
                    )
                )

    @property
    def ctx(self) -> Context:
        """The context this texture belongs to."""
        return self._ctx

    @property
    def glo(self) -> gl.GLuint:
        """The OpenGL texture id"""
        return self._glo

    @property
    def compressed(self) -> bool:
        """Is this using a compressed format?"""
        return self._compressed

    @property
    def width(self) -> int:
        """The width of the texture in pixels"""
        return self._width

    @property
    def height(self) -> int:
        """The height of the texture in pixels"""
        return self._height

    @property
    def dtype(self) -> str:
        """The data type of each component"""
        return self._dtype

    @property
    def size(self) -> tuple[int, int]:
        """The size of the texture as a tuple"""
        return self._width, self._height

    @property
    def samples(self) -> int:
        """Number of samples if multisampling is enabled (read only)"""
        return self._samples

    @property
    def byte_size(self) -> int:
        """The byte size of the texture."""
        return pixel_formats[self._dtype][3] * self._components * self.width * self.height

    @property
    def components(self) -> int:
        """Number of components in the texture"""
        return self._components

    @property
    def component_size(self) -> int:
        """Size in bytes of each component"""
        return self._component_size

    @property
    def depth(self) -> bool:
        """If this is a depth texture."""
        return self._depth

    @property
    def immutable(self) -> bool:
        """Does this texture have immutable storage?"""
        return self._immutable

    @property
    def swizzle(self) -> str:
        """
        The swizzle mask of the texture (Default ``'RGBA'``).

        The swizzle mask change/reorder the ``vec4`` value returned by the ``texture()`` function
        in a GLSL shaders. This is represented by a 4 character string were each
        character can be::

            'R' GL_RED
            'G' GL_GREEN
            'B' GL_BLUE
            'A' GL_ALPHA
            '0' GL_ZERO
            '1' GL_ONE

        Example::

            # Alpha channel will always return 1.0
            texture.swizzle = 'RGB1'

            # Only return the red component. The rest is masked to 0.0
            texture.swizzle = 'R000'

            # Reverse the components
            texture.swizzle = 'ABGR'
        """
        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)

        # Read the current swizzle values from the texture
        swizzle_r = gl.GLint()
        swizzle_g = gl.GLint()
        swizzle_b = gl.GLint()
        swizzle_a = gl.GLint()

        gl.glGetTexParameteriv(self._target, gl.GL_TEXTURE_SWIZZLE_R, swizzle_r)
        gl.glGetTexParameteriv(self._target, gl.GL_TEXTURE_SWIZZLE_G, swizzle_g)
        gl.glGetTexParameteriv(self._target, gl.GL_TEXTURE_SWIZZLE_B, swizzle_b)
        gl.glGetTexParameteriv(self._target, gl.GL_TEXTURE_SWIZZLE_A, swizzle_a)

        swizzle_str = ""
        for v in [swizzle_r, swizzle_g, swizzle_b, swizzle_a]:
            swizzle_str += swizzle_enum_to_str[v.value]

        return swizzle_str

    @swizzle.setter
    def swizzle(self, value: str):
        if not isinstance(value, str):
            raise ValueError(f"Swizzle must be a string, not '{type(str)}'")

        if len(value) != 4:
            raise ValueError("Swizzle must be a string of length 4")

        swizzle_enums = []
        for c in value:
            try:
                c = c.upper()
                swizzle_enums.append(swizzle_str_to_enum[c])
            except KeyError:
                raise ValueError(f"Swizzle value '{c}' invalid. Must be one of RGBA01")

        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)

        gl.glTexParameteri(self._target, gl.GL_TEXTURE_SWIZZLE_R, swizzle_enums[0])
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_SWIZZLE_G, swizzle_enums[1])
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_SWIZZLE_B, swizzle_enums[2])
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_SWIZZLE_A, swizzle_enums[3])

    @property
    def filter(self) -> tuple[int, int]:
        """
        Get or set the ``(min, mag)`` filter for this texture.

        These are rules for how a texture interpolates.
        The filter is specified for minification and magnification.

        Default value is ``LINEAR, LINEAR``.
        Can be set to ``NEAREST, NEAREST`` for pixelated graphics.

        When mipmapping is used the min filter needs to be one of the
        ``MIPMAP`` variants.

        Accepted values::

            # Enums can be accessed on the context or arcade.gl
            NEAREST                # Nearest pixel
            LINEAR                 # Linear interpolate
            NEAREST_MIPMAP_NEAREST # Minification filter for mipmaps
            LINEAR_MIPMAP_NEAREST  # Minification filter for mipmaps
            NEAREST_MIPMAP_LINEAR  # Minification filter for mipmaps
            LINEAR_MIPMAP_LINEAR   # Minification filter for mipmaps

        Also see

        * https://www.khronos.org/opengl/wiki/Texture#Mip_maps
        * https://www.khronos.org/opengl/wiki/Sampler_Object#Filtering
        """
        return self._filter

    @filter.setter
    def filter(self, value: tuple[int, int]):
        if not isinstance(value, tuple) or not len(value) == 2:
            raise ValueError("Texture filter must be a 2 component tuple (min, mag)")

        self._filter = value
        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_MIN_FILTER, self._filter[0])
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_MAG_FILTER, self._filter[1])

    @property
    def wrap_x(self) -> int:
        """
        Get or set the horizontal wrapping of the texture.

        This decides how textures are read when texture coordinates are outside
        the ``[0.0, 1.0]`` area. Default value is ``REPEAT``.

        Valid options are::

            # Note: Enums can also be accessed in arcade.gl
            # Repeat pixels on the y axis
            texture.wrap_x = ctx.REPEAT
            # Repeat pixels on the y axis mirrored
            texture.wrap_x = ctx.MIRRORED_REPEAT
            # Repeat the edge pixels when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_EDGE
            # Use the border color (black by default) when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_BORDER
        """
        return self._wrap_x

    @wrap_x.setter
    def wrap_x(self, value: int):
        self._wrap_x = value
        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_WRAP_S, value)

    @property
    def wrap_y(self) -> int:
        """
        Get or set the horizontal wrapping of the texture.

        This decides how textures are read when texture coordinates are outside the
        ``[0.0, 1.0]`` area. Default value is ``REPEAT``.

        Valid options are::

            # Note: Enums can also be accessed in arcade.gl
            # Repeat pixels on the x axis
            texture.wrap_x = ctx.REPEAT
            # Repeat pixels on the x axis mirrored
            texture.wrap_x = ctx.MIRRORED_REPEAT
            # Repeat the edge pixels when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_EDGE
            # Use the border color (black by default) when reading outside the texture
            texture.wrap_x = ctx.CLAMP_TO_BORDER
        """
        return self._wrap_y

    @wrap_y.setter
    def wrap_y(self, value: int):
        self._wrap_y = value
        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)
        gl.glTexParameteri(self._target, gl.GL_TEXTURE_WRAP_T, value)

    @property
    def anisotropy(self) -> float:
        """Get or set the anisotropy for this texture."""
        return self._anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self._anisotropy = max(1.0, min(value, self._ctx.info.MAX_TEXTURE_MAX_ANISOTROPY))
        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)
        gl.glTexParameterf(self._target, gl.GL_TEXTURE_MAX_ANISOTROPY, self._anisotropy)

    @property
    def compare_func(self) -> str | None:
        """
        Get or set the compare function for a depth texture::

            texture.compare_func = None  # Disable depth comparison completely
            texture.compare_func = '<='  # GL_LEQUAL
            texture.compare_func = '<'   # GL_LESS
            texture.compare_func = '>='  # GL_GEQUAL
            texture.compare_func = '>'   # GL_GREATER
            texture.compare_func = '=='  # GL_EQUAL
            texture.compare_func = '!='  # GL_NOTEQUAL
            texture.compare_func = '0'   # GL_NEVER
            texture.compare_func = '1'   # GL_ALWAYS
        """
        return self._compare_func

    @compare_func.setter
    def compare_func(self, value: str | None):
        if not self._depth:
            raise ValueError("Depth comparison function can only be set on depth textures")

        if not isinstance(value, str) and value is not None:
            raise ValueError(f"value must be as string: {self._compare_funcs.keys()}")

        func = compare_funcs.get(value, None)
        if func is None:
            raise ValueError(f"value must be as string: {compare_funcs.keys()}")

        self._compare_func = value
        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(self._target, self._glo)
        if value is None:
            gl.glTexParameteri(self._target, gl.GL_TEXTURE_COMPARE_MODE, gl.GL_NONE)
        else:
            gl.glTexParameteri(
                self._target, gl.GL_TEXTURE_COMPARE_MODE, gl.GL_COMPARE_REF_TO_TEXTURE
            )
            gl.glTexParameteri(self._target, gl.GL_TEXTURE_COMPARE_FUNC, func)

    def read(self, level: int = 0, alignment: int = 1) -> bytes:
        """
        Read the contents of the texture.

        Args:
            level:
                The texture level to read
            alignment:
                Alignment of the start of each row in memory in number of bytes.
                Possible values: 1,2,4
        """
        if self._samples > 0:
            raise ValueError("Multisampled textures cannot be read directly")

        if self._ctx.gl_api == "gl":
            gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
            gl.glBindTexture(self._target, self._glo)
            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, alignment)

            buffer = (
                gl.GLubyte * (self.width * self.height * self._component_size * self._components)
            )()
            gl.glGetTexImage(gl.GL_TEXTURE_2D, level, self._format, self._type, buffer)
            return string_at(buffer, len(buffer))
        elif self._ctx.gl_api == "gles":
            fbo = self._ctx.framebuffer(color_attachments=[self])
            return fbo.read(components=self._components, dtype=self._dtype)
        else:
            raise ValueError("Unknown gl_api: '{self._ctx.gl_api}'")

    def write(self, data: BufferOrBufferProtocol, level: int = 0, viewport=None) -> None:
        """Write byte data from the passed source to the texture.

        The ``data`` value can be either an
        :py:class:`arcade.gl.Buffer` or anything that implements the
        `Buffer Protocol <https://docs.python.org/3/c-api/buffer.html>`_.

        The latter category includes ``bytes``, ``bytearray``,
        ``array.array``, and more. You may need to use typing
        workarounds for non-builtin types. See
        :ref:`prog-guide-gl-buffer-protocol-typing` for more
        information.

        Args:
            data:
                :class:`~arcade.gl.Buffer` or buffer protocol object with data to write.
            level:
                The texture level to write
            viewport:
                The area of the texture to write. 2 or 4 component tuple.
                (x, y, w, h) or (w, h). Default is the full texture.
        """
        # TODO: Support writing to layers using viewport + alignment
        if self._samples > 0:
            raise ValueError("Writing to multisampled textures not supported")

        x, y, w, h = 0, 0, self._width, self._height
        if viewport:
            if len(viewport) == 2:
                w, h = viewport
            elif len(viewport) == 4:
                x, y, w, h = viewport
            else:
                raise ValueError("Viewport must be of length 2 or 4")

        if isinstance(data, Buffer):
            gl.glBindBuffer(gl.GL_PIXEL_UNPACK_BUFFER, data.glo)
            gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
            gl.glBindTexture(self._target, self._glo)
            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, 1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
            gl.glTexSubImage2D(self._target, level, x, y, w, h, self._format, self._type, 0)
            gl.glBindBuffer(gl.GL_PIXEL_UNPACK_BUFFER, 0)
        else:
            byte_size, data = data_to_ctypes(data)
            self._validate_data_size(data, byte_size, w, h)
            gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
            gl.glBindTexture(self._target, self._glo)
            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, 1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
            gl.glTexSubImage2D(
                self._target,  # target
                level,  # level
                x,  # x offset
                y,  # y offset
                w,  # width
                h,  # height
                self._format,  # format
                self._type,  # type
                data,  # pixel data
            )

    def _validate_data_size(self, byte_data, byte_size, width, height) -> None:
        """Validate the size of the data to be written to the texture"""
        # TODO: Validate data size for compressed textures
        #       This might be a bit tricky since the size of the compressed
        #       data would depend on the algorithm used.
        if self._compressed is True:
            return

        expected_size = width * height * self._component_size * self._components
        if byte_size != expected_size:
            raise ValueError(
                f"Data size {len(byte_data)} does not match expected size {expected_size}"
            )
        if len(byte_data) != byte_size:
            raise ValueError(
                f"Data size {len(byte_data)} does not match reported size {expected_size}"
            )

    def build_mipmaps(self, base: int = 0, max_level: int = 1000) -> None:
        """Generate mipmaps for this texture.

        The default values usually work well.

        Mipmaps are successively smaller versions of an original
        texture with special filtering applied. Using mipmaps allows
        OpenGL to render scaled versions of original textures with fewer
        scaling artifacts.

        Mipmaps can be made for textures of any size. Each mipmap
        version halves the width and height of the previous one (e.g.
        256 x 256, 128 x 128, 64 x 64, etc) down to a minimum of 1 x 1.

        .. note:: Mipmaps will only be used if a texture's filter is
                  configured with a mipmap-type minification::

                   # Set up linear interpolating minification filter
                   texture.filter = ctx.LINEAR_MIPMAP_LINEAR, ctx.LINEAR

        Args:
            base:
                Level the mipmaps start at (usually 0)
            max_level:
                The maximum number of levels to generate

        Also see: https://www.khronos.org/opengl/wiki/Texture#Mip_maps
        """
        if self._samples > 0:
            raise ValueError("Multisampled textures don't support mimpmaps")

        gl.glActiveTexture(gl.GL_TEXTURE0 + self._ctx.default_texture_unit)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._glo)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_BASE_LEVEL, base)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAX_LEVEL, max_level)
        gl.glGenerateMipmap(gl.GL_TEXTURE_2D)

    def delete(self):
        """
        Destroy the underlying OpenGL resource.

        Don't use this unless you know exactly what you are doing.
        """
        Texture2D.delete_glo(self._ctx, self._glo)
        self._glo.value = 0

    @staticmethod
    def delete_glo(ctx: "Context", glo: gl.GLuint):
        """
        Destroy the texture.

        This is called automatically when the object is garbage collected.

        Args:
            ctx: OpenGL Context
            glo: The OpenGL texture id
        """
        # If we have no context, then we are shutting down, so skip this
        if gl.current_context is None:
            return

        if glo.value != 0:
            gl.glDeleteTextures(1, byref(glo))

        ctx.stats.decr("texture")

    def use(self, unit: int = 0) -> None:
        """Bind the texture to a channel,

        Args:
            unit: The texture unit to bind the texture.
        """
        gl.glActiveTexture(gl.GL_TEXTURE0 + unit)
        gl.glBindTexture(self._target, self._glo)

    def bind_to_image(self, unit: int, read: bool = True, write: bool = True, level: int = 0):
        """
        Bind textures to image units.

        Note that either or both ``read`` and ``write`` needs to be ``True``.
        The supported modes are: read only, write only, read-write

        Args:
            unit: The image unit
            read: The compute shader intends to read from this image
            write: The compute shader intends to write to this image
            level: The mipmap level to bind
        """
        if self._ctx.gl_api == "gles" and not self._immutable:
            raise ValueError("Textures bound to image units must be created with immutable=True")

        access = gl.GL_READ_WRITE
        if read and write:
            access = gl.GL_READ_WRITE
        elif read and not write:
            access = gl.GL_READ_ONLY
        elif not read and write:
            access = gl.GL_WRITE_ONLY
        else:
            raise ValueError("Illegal access mode. The texture must at least be read or write only")

        gl.glBindImageTexture(unit, self._glo, level, 0, 0, access, self._internal_format)

    def get_handle(self, resident: bool = True) -> int:
        """
        Get a handle for bindless texture access.

        Once a handle is created its parameters cannot be changed.
        Attempting to do so will have no effect. (filter, wrap etc).
        There is no way to undo this immutability.

        Handles cannot be used by shaders until they are resident.
        This method can be called multiple times to move a texture
        in and out of residency::

            >> texture.get_handle(resident=False)
            4294969856
            >> texture.get_handle(resident=True)
            4294969856

        Ths same handle is returned if the handle already exists.

        .. note:: Limitations from the OpenGL wiki

            The amount of storage available for resident images/textures may be less
            than the total storage for textures that is available. As such, you should
            attempt to minimize the time a texture spends being resident. Do not attempt
            to take steps like making textures resident/un-resident every frame or something.
            But if you are finished using a texture for some time, make it un-resident.

        Args:
            resident: Make the texture resident.
        """
        handle = gl.glGetTextureHandleARB(self._glo)
        is_resident = gl.glIsTextureHandleResidentARB(handle)

        # Ensure we don't try to make a resident texture resident again
        if resident:
            if not is_resident:
                gl.glMakeTextureHandleResidentARB(handle)
        else:
            if is_resident:
                gl.glMakeTextureHandleNonResidentARB(handle)

        return handle

    def __repr__(self) -> str:
        return "<Texture glo={} size={}x{} components={}>".format(
            self._glo.value, self._width, self._height, self._components
        )



from dataclasses import dataclass
from typing import TypeAlias

import arcade
from arcade import Texture, color, uicolor
from arcade.gui.nine_patch import NinePatchTexture
from arcade.gui.property import DictProperty, bind
from arcade.gui.style import UIStyleBase, UIStyledWidget
from arcade.gui.surface import Surface
from arcade.gui.widgets import UIInteractiveWidget
from arcade.gui.widgets.text import UITextWidget
from arcade.text import FontNameOrNames
from arcade.types import RGBA255


@dataclass
class UITextureButtonStyle(UIStyleBase):
    """Used to style the texture button. Below is its use case.

    .. code:: py

        button = UITextureButton(style={"normal": UITextureButton.UIStyle(...),})
    """

    font_size: int = 12
    font_name: FontNameOrNames = ("Kenney Future", "arial", "calibri")
    font_color: RGBA255 = uicolor.WHITE


class UITextureButton(UIInteractiveWidget, UIStyledWidget[UITextureButtonStyle], UITextWidget):
    """A button with an image for the face of the button.

    There are four states of the UITextureButton i.e.normal, hovered, pressed and disabled.

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget. Defaults to texture width if not
            specified.
        height: height of widget. Defaults to texture height if not
            specified.
        texture: texture to display for the widget.
        texture_hovered: different texture to display if mouse is
            hovering over button.
        texture_pressed: different texture to display if mouse button is
            pressed while hovering over button.
        text: text to add to the button.
        multiline: allows to wrap text, if not enough width available
        style: Used to style the button for different states.
        scale: scale the button, based on the base texture size.
        size_hint: Tuple of floats (0.0-1.0), how much space of the
            parent should be requested
        size_hint_min: min width and height in pixel
        size_hint_max: max width and height in pixel
    """

    _textures = DictProperty[str, Texture | NinePatchTexture]()

    UIStyle = UITextureButtonStyle

    DEFAULT_STYLE = {
        "normal": UIStyle(),
        "hover": UIStyle(
            font_color=uicolor.WHITE_CLOUDS,
        ),
        "press": UIStyle(
            font_color=uicolor.DARK_BLUE_MIDNIGHT_BLUE,
        ),
        "disabled": UIStyle(
            font_color=uicolor.WHITE_SILVER,
        ),
    }

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float | None = None,
        height: float | None = None,
        texture: None | Texture | NinePatchTexture = None,
        texture_hovered: None | Texture | NinePatchTexture = None,
        texture_pressed: None | Texture | NinePatchTexture = None,
        texture_disabled: None | Texture | NinePatchTexture = None,
        text: str = "",
        multiline: bool = False,
        scale: float | None = None,
        style: dict[str, UIStyleBase] | None = None,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        **kwargs,
    ):
        if width is None and texture is not None:
            width = texture.size[0]

        if height is None and texture is not None:
            height = texture.size[1]

        if width is None:
            raise ValueError("Unable to determine a width.")
        if height is None:
            raise ValueError("Unable to determine a height.")

        if scale is not None and texture is not None:
            width = texture.size[0] * scale
            height = texture.size[1] * scale

        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            style=style or self.DEFAULT_STYLE,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            text=text,
            multiline=multiline,
            **kwargs,
        )

        self._textures = {}

        if texture:
            self._textures["normal"] = texture
            self._textures["hover"] = texture
            self._textures["press"] = texture
            self._textures["disabled"] = texture
        if texture_hovered:
            self._textures["hover"] = texture_hovered
        if texture_pressed:
            self._textures["press"] = texture_pressed
        if texture_disabled:
            self._textures["disabled"] = texture_disabled

        bind(self, "_textures", self.trigger_render)

        # prepare label with default style
        _style = self.get_current_style()
        if _style is None:
            raise ValueError(f"No style found for state {self.get_current_state()}")
        self._apply_style(_style)

    def get_current_state(self) -> str:
        """Returns the current state of the button i.e.disabled, press, hover or normal."""
        if self.disabled:
            return "disabled"
        elif self.pressed:
            return "press"
        elif self.hovered or self.focused:
            return "hover"
        else:
            return "normal"

    @property
    def texture(self):
        """Returns the normal texture for the face of the button."""
        return self._textures["normal"]

    @texture.setter
    def texture(self, value: Texture):
        self._textures["normal"] = value
        self.trigger_render()

    @property
    def texture_hovered(self):
        """Returns the hover texture for the face of the button."""
        return self._textures["hover"]

    @texture_hovered.setter
    def texture_hovered(self, value: Texture):
        self._textures["hover"] = value
        self.trigger_render()

    @property
    def texture_pressed(self):
        """Returns the pressed texture for the face of the button."""
        return self._textures["press"]

    @texture_pressed.setter
    def texture_pressed(self, value: Texture):
        self._textures["press"] = value
        self.trigger_render()

    def do_render(self, surface: Surface):
        """Render the widgets graphical representation."""
        self.prepare_render(surface)

        style = self.get_current_style()

        # update label
        if style is None:
            raise ValueError(f"No style found for state {self.get_current_state()}")
        self._apply_style(style)

        current_state = self.get_current_state()
        current_texture = self._textures.get(current_state)
        if current_texture:
            surface.draw_texture(0, 0, self.content_width, self.content_height, current_texture)

    def _apply_style(self, style: UITextureButtonStyle):
        """Callback which is called right before rendering to apply changes for rendering."""
        font_name = style.get("font_name", UIFlatButton.UIStyle.font_name)
        font_size = style.get("font_size", UIFlatButton.UIStyle.font_size)
        font_color = style.get("font_color", UIFlatButton.UIStyle.font_color)
        self.ui_label.update_font(font_name, font_size, font_color)

        self.ui_label.rect = self.ui_label.rect.max_size(self.content_width, self.content_height)


@dataclass
class UIFlatButtonStyle(UIStyleBase):
    """Used to style the button. Below is its use case.

    .. code:: py

        button = UIFlatButton(style={"normal": UIFlatButton.UIStyle(...),})
    """

    font_size: int = 12
    font_name: FontNameOrNames = ("Kenney Future", "arial", "calibri")
    font_color: RGBA255 = color.WHITE
    bg: RGBA255 = uicolor.DARK_BLUE_MIDNIGHT_BLUE
    border: RGBA255 | None = None
    border_width: int = 0


class UIFlatButton(UIInteractiveWidget, UIStyledWidget[UIFlatButtonStyle], UITextWidget):
    """A text button, with support for background color and a border.

    There are four states of the UITextureButton i.e. normal, hovered, pressed and disabled.

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget. Defaults to texture width if not
            specified.
        height: height of widget. Defaults to texture height if not
            specified.
        text: text to add to the button.
        multiline: allows to wrap text, if not enough width available
        style: Used to style the button
    """

    UIStyle: TypeAlias = UIFlatButtonStyle

    DEFAULT_STYLE = {
        "normal": UIStyle(),
        "hover": UIStyle(
            font_color=color.WHITE,
            bg=uicolor.DARK_BLUE_WET_ASPHALT,
            border=uicolor.GRAY_CONCRETE,
        ),
        "press": UIStyle(
            font_color=uicolor.DARK_BLUE_MIDNIGHT_BLUE,
            bg=uicolor.WHITE_CLOUDS,
            border=uicolor.GRAY_CONCRETE,
        ),
        "disabled": UIStyle(
            font_color=uicolor.WHITE_SILVER,
            bg=uicolor.GRAY_ASBESTOS,
        ),
    }

    STYLE_RED = {
        "normal": UIStyle(
            font_color=uicolor.WHITE_CLOUDS,
            bg=uicolor.RED_ALIZARIN,
            border=uicolor.RED_POMEGRANATE,
        ),
        "hover": UIStyle(
            bg=uicolor.RED_ALIZARIN,
            font_color=uicolor.WHITE_CLOUDS,
            border=uicolor.WHITE_SILVER,
            border_width=2,
        ),
        "press": UIStyle(
            bg=uicolor.RED_POMEGRANATE,
            font_color=uicolor.WHITE_CLOUDS,
            border=uicolor.WHITE_SILVER,
            border_width=2,
        ),
        "disabled": UIStyle(
            bg=uicolor.GRAY_ASBESTOS,
        ),
    }

    STYLE_BLUE = {
        "normal": UIStyle(bg=uicolor.BLUE_PETER_RIVER, font_color=uicolor.WHITE_CLOUDS),
        "hover": UIStyle(
            bg=uicolor.BLUE_BELIZE_HOLE,
            font_color=uicolor.WHITE_CLOUDS,
            border=uicolor.WHITE_SILVER,
            border_width=2,
        ),
        "press": UIStyle(
            bg=uicolor.DARK_BLUE_MIDNIGHT_BLUE,
            font_color=uicolor.WHITE_CLOUDS,
            border=uicolor.WHITE_SILVER,
            border_width=2,
        ),
        "disabled": UIStyle(bg=uicolor.GRAY_ASBESTOS),
    }

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 100,
        height: float = 50,
        text="",
        multiline=False,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        style=None,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            style=style or self.DEFAULT_STYLE,
            text=text,
            multiline=multiline,
            **kwargs,
        )

        # prepare label with default style
        _style = self.get_current_style()
        if _style is None:
            raise ValueError(f"No style found for state {self.get_current_state()}")
        self._apply_style(_style)

    def get_current_state(self) -> str:
        """Returns the current state of the button i.e.disabled, press, hover or normal."""
        if self.disabled:
            return "disabled"
        elif self.pressed:
            return "press"
        elif self.hovered or self.focused:
            return "hover"
        else:
            return "normal"

    def do_render(self, surface: Surface):
        """Render a flat button, graphical representation depends on the current state."""
        self.prepare_render(surface)
        style = self.get_current_style()
        if style is None:
            raise ValueError(f"No style found for state {self.get_current_state()}")

        # update label
        # this might trigger another render run, due to label size change
        self._apply_style(style)

        # Render button
        border_width = style.get("border_width", UIFlatButton.UIStyle.border_width)
        border_color = style.get("border", UIFlatButton.UIStyle.border)
        bg_color = style.get("bg", UIFlatButton.UIStyle.bg)
        if bg_color:
            surface.clear(bg_color)

        # render button border (which is not the widgets border)
        if border_color and border_width:
            arcade.draw_lbwh_rectangle_outline(
                border_width,
                border_width,
                self.content_width - 2 * border_width,
                self.content_height - 2 * border_width,
                color=border_color,
                border_width=border_width,
            )

    def _apply_style(self, style: UIStyle):
        """Callback which is called right before rendering to apply changes for rendering."""
        font_name = style.get("font_name", UIFlatButton.UIStyle.font_name)
        font_size = style.get("font_size", UIFlatButton.UIStyle.font_size)
        font_color = style.get("font_color", UIFlatButton.UIStyle.font_color)
        self.ui_label.update_font(font_name, font_size, font_color)
